"""
A parser/processor to simplify modular docopt cli
"""
import logging
import os
import re
import sys
from getpass import getpass
from textwrap import dedent

from docopt import docopt, DocoptExit, DocoptLanguageError


class CommandParser:
    """
    A pythonic, declarative and modular approach to cli (based on docopt)

    Why?
        To make cli implementation a breeze, perhaps even fun: No more untestable
        if/then mess. Every command is a class, every action is a method.

    Tell me more!
        CommandParser provides a straight-forward way to implement a modular
        cli, using docopt as the command line parser. In a nutshell,
        a Command is a Python class that provides a method for each action
        to be performed, thus greatly reducing the need for if/then constructs
        while modularising documentation and implementation of subcommands.

    Why not use docopt directly?
        doctopt is a great way to specify and parse command line arguments
        and CommandParser uses it as the actual parser. With docopt however
        the developer is left to provide an interpretation and implementation
        of the parsed arguments. The implementation quickly becomes a large
        cascade of if/then/else: a bug feast waiting to happen and a
        maintenance nightmare. Also hard to write unit tests. Note that's not
        the fault of docopt, it's does what it does flawlessly. CommandParser
        provides the missing link from docopt's parsing to implementation.


    Usage:
        # pgm.py
        ___doc___ = '''Usage: pgm.py foo bar [--baz]
                                     foo bax [--pox]

        [usage:foo]

        Options:
            -h | --help  this help text

        [options:foo]
        '''
        CommandParser(__doc__, [FooCommand, ...]).parse().process()
        # that's it, notice there are no if/then cascades
        # end of pgm.py

        where FooCommand is

            # foo.py
            class FooCommand(CommandBase):
                '''
                Usage: pgm.py foo bar [--baz]

                Options:
                    --baz  baz description
                '''
                command = 'foo'
                usage_header = 'Working with foo' # optional
                options_header = 'Options for foo' # optional

                def bar(self):
                    # the bar command, docopt output is in self.args
                    # as parsed to the docstring of foo.py. If you want
                    # self.args of pgm.py, access self.parser.args
                    #
                    # to write output, use self.logger, a Python logger
                    # which by default prints to stdout
                    ... your code ...
                    baz = self.args.get('--baz')
                    self.logger.info(f"Hello {baz}")

                def bax(self):
                    # the bax command ...
                    pox = self.args.get('--pox')
                    self.logger.info(f"Welcome {pox}")

                # notice the absence of if/then. every action is a method
            # -- end of foo.py

        Upon the call to .process(), CommandParser will find and call the
        FooCommmand.bar() method automatically. In bar() use self.args
        to get the parsed arguments as returned by docopt(FooCommand.__doc__).

        CommandParser will automatically replace [usage:foo] and [options:foo]
        with the FooCommand.__doc__ string, replacing the latter's Usage:
        with the usage_header and Options: with options_header

        The help text will thus be:

            Usage: pgm.py foo bar [--baz]

            Working with foo
                pgm.py foo bar [--baz]

            Options:
                -h | --help  this help text

            Options for foo
                --baz  baz description

    How to implement help:
        CommandParser automatically handles -h, --help and prog help <action>
        -h prints only the usage section, --help and prog help will print all sections

        If a command is given (e.g. prog foo -h) it will print the usage or help
        of the command.

        To override what is printed for help, implement the Command.help() method.

            # FooCommand
            def help(self, usage_only=True):
                ...

    How to print output in a command:
        You should avoid using print() as it is not really a good way to
        write modular, debuggable code. If your cli is more than a few
        lines of code, it is better to use the built-in logger:

            (...)
            def bar(self):
                self.logger.info('this will go to stdout')

        To change the format of the output, create a logging.Logger and
        set whatever format you like, then use

            logger = ... # your logger setup
            CommandParser(..., logger=logger)

        To change the loglevel, add --loglevel in options, pass INFO, DEBUG
        or ERROR


    Optional Description tag:
        For the top level module and each command you can add more descriptions
        that only get printed on prog help or prog help <action>

        # pgm.py
        (...)

        Description:
            lorem ipsum lorem ipsum lorem ipsum lorem ipsum lorem ipsum lorem ipsum
            lorem ipsum lorem ipsum lorem ipsum lorem ipsum lorem ipsum lorem ipsum

        [description:foo]

        # foo.py
        '''
        (...)

        Description:
            lorem ipsum lorem ipsum lorem ipsum lorem ipsum lorem ipsum lorem ipsum
            lorem ipsum lorem ipsum lorem ipsum lorem ipsum lorem ipsum lorem ipsum
        '''

        The help text will thus be:

            Usage: pgm.py foo bar [--baz]

            Working with foo
                pgm.py foo bar [--baz]

            Options:
                -h | --help  this help text

            Options for foo
                --baz  baz description

            Description for foo
                lorem ipsum lorem ipsum lorem ipsum lorem ipsum lorem ipsum lorem ipsum
                lorem ipsum lorem ipsum lorem ipsum lorem ipsum lorem ipsum lorem ipsum


    My Command's parameters are not recognized:
        Before a command's docopts are parsed, the top-level docopt must parse.
        Thus if you do something like this, your command will never receive
        a call:

        # pgm.py
        '''
        Usage: pgm foo
        '''

        # foo.py
        '''
        Usage: pgm foo [--baz]
        '''

        To make it work, add [options:foo] to pgm.py, and either

        (a) add an Options: section in foo.py that includes --baz,
        (b) or include [--baz] as a global option

        Note the effect of (a) and (b) is effectively the same, as every
        command's options section is added to the global Options: if
        there is a corresponding [foo:<command>] placeholder.


    Debugging:
        To debug the parser, set DOCOPT_DEBUG=1. This will print the parsed
        arguments in self.args, along with the command class and method chosen
        for execution on callling parser.process().

        If you don't see debug output despite specifying DOCOPT_DEBUG=1 the
        parser has not been able to parse the arguments and instead prints the
        help text. This means that your argv don't match any of the
        specified docopts. Make sure that your top-level docstring supports
        all arguments and options.

        The command class is determined by a match to CommandBase.command,
        in this order:
            1. a valid literal passed in argv
            2. the value of the <command> tag

        The method called on the command class is determined in this order:
            3. a valid literal passed in argv
            4. the value of the <action> tag
            5. __call__()

        For example, if you specify (1) or (2)

            (1)
            pgm.py foo bar

            (2)
            pgm.py <command> <action>

        (1) argv = "pgm.py foo bar" => if there is FooCommand.command == 'foo', the
            "foo" part is a literal match. The "bar" part is also a literal match.

        (2) argv = "pgm.py foo bar" => if there is FooCommand.command == 'foo', the
            "foo" part is a match by the <command> tag. The "bar" part is a match
            by the "<action>" tag. It is translated to the corresponding method of
            the same name if it exists, or reverts to FooCommand.__call()

        In both (1) and (2) the "bar" part is translated to calling FooCommand.bar()
        if it exists, else reverts to FooCommand.__call_().

        The default command tag is '<command>', change by setting CommandParser.command_tag
        to something else:

            parser = CommandParser(..., command_tag='<dothis>')

        The default action tag is '<action>'. Change by setting the Command.action_tag
        to something else.

            class FooCommand(BaseCommand):
                action_tag = '<dothis>'
                ...

    """

    def __init__(self, docs, commands, argv=None, version=None, logger=None,
                 command_tag=None, catchall='catchall', askfn=None):
        self.docs = docs
        self.commands = commands  # available command implementations
        self.command = None  # chosen command implementation, see parse_command
        self.command_tag = command_tag or '<command>'
        self.argv = argv or sys.argv[1:]
        self.version = version
        self.args = None
        self.catchall = catchall
        self._logger = logger
        self._askfn = askfn
        self.apply_command_usage()

    @property
    def logger(self):
        """
        return the logger instance, defaults to the console logger

        Returns:
            a logger, any command instance should use this logger for output
            this helps not only output control but also testing.
        """
        if self._logger is None:
            self._logger = setup_console_logger()
            loglevel = self.args.get('--loglevel', 'INFO')
            self._logger.setLevel(loglevel)
        return self._logger

    @property
    def should_debug(self):
        return os.environ.get('DOCOPT_DEBUG')

    @property
    def silent(self):
        return self.args.get('-q') or self.args.get('--noinput')

    def apply_command_usage(self):
        """
        For every command add usage, options and descriptions to global docs

        This works by extracting the usage, options and descriptions from
        the doc string of every registered command class and adding these
        to the global docs by replacing the corresponding [usage:<command>],
        [options:<command>] and [descriptions:<command] sections, respectively.

        Returns:
            None
        """
        # replace [usage:<command>] [options:<command>] placeholders with command.__doc__
        for command in self.commands:
            if command.__doc__:
                usage, options, description = command.get_command_docparts()
                usage_placeholder = '[usage:{}]'.format(command.command)
                options_placeholder = '[options:{}]'.format(command.command)
                descr_placeholder = '[description:{}]'.format(command.command)
                self.docs = self.docs.replace(usage_placeholder, usage)
                self.docs = self.docs.replace(options_placeholder, options)
                self.docs = re.sub(r'\n\s*\n', '\n\n', self.docs)
                if any(v in self.argv for v in ('--help-ext', '-H', '-hh', 'help')):
                    self.docs = self.docs.replace(descr_placeholder, description)
                else:
                    self.docs = self.docs.replace(descr_placeholder, '')

    def parse(self):
        """
        Parse the doc string given on instantiation, determine current command

        Returns:
            the parser self
        """
        try:
            self.args = safe_docopt(self.docs, argv=self.argv, version=self.version, help=False)
        except DocoptExit:
            # by specifying docopt(help=False) we get to control what happens on parse failure
            if self.argv:
                # if help requested, show help. this simulates a valid Usage: help <action>
                if self.argv[0] in ('help', '--help'):
                    action = self.argv[1] if len(self.argv) > 1 else None
                    self.args = {
                        '<command>': 'help',
                        '<action>': action,
                    }
                    self.help(usage_only=False)
                    # raise SystemExit because
                    raise SystemExit
                # custom handling should be provided by a Command with command='catchall'
                elif self.argv[0] == '--copyright':
                    self.args = {
                        '--copyright': True,
                    }
                else:
                    # this re-raises DocoptExit which prints usage
                    raise
            else:
                # this re-raises DocoptExit which prints usage
                raise
        # if we get here it means docopt has decided all the arguments are valid
        # -- see if we have a command
        self.command = self.parse_command()
        # -- no command?!
        if not self.command:
            self.help()
            raise DocoptExit()
        if self.should_debug:
            print("*** docopt parsed args", self.args)
            print("*** docopt using command class {}".format(repr(self.command)))
            print("*** docopt using command method {}".format(repr(self.command.get_command_method())))
        return self

    def parse_command(self):
        """
        Instantiate the command requested

        The command requested is determined by considering
            * a literal as specified
            * a <command> placeholder
            * neither literal nor <command>, the 'catchall' command, if present

        Returns:
            the command class or None
        """
        for commandcls in self.commands:
            is_global_command = False
            # given as literal
            command_asliteral = self.args.get(commandcls.command)
            # given as <command> arg
            command_asarg = self.args.get(self.command_tag) == commandcls.command
            # catch all?
            if not (command_asliteral or command_asarg):
                is_global_command = commandcls.command == self.catchall
            if command_asliteral or command_asarg or is_global_command:
                command = self._command_instance(commandcls)
                command.parse()
                return command
        self.help()

    def _command_instance(self, commandcls):
        # instantiate the commandcls given current arguments and a parsed global instance
        command = commandcls(commandcls.__doc__, argv=self.argv,
                             logger=self.logger, parser=self)
        return command

    def process(self):
        """
        Process the command as identified in parse

        Processing means to find the <action> method to call on the command class,
        see get_command_method() for details

        Returns:

        """
        assert self.command is not None, "you must call CommandParser.parse()"
        meth = self.command.get_command_method()
        if meth:
            return meth()

    def help(self, usage_only=True):
        """
        show default help

        This assumes that the docs contain any of the following Usage patterns:

            Usage:
                prog <command> <action>
                prog help <a-command>

        Either of the following is shown, in this order:
            - if the command was 'help', but no <a-command> was given, show
              self.docs
            - if <a-command> was given this is used to lookup the actual command,
              then calls command.help()
            - if <command> was not 'help', i.e. we got here by some command
              calling parser.help(), we show usage information by calling the
              command's command.help(usage_only=True)
            - if no command can be identified, we let the user know we don't
              know what to do. If possible we hint at the closest command that
              is similar to the requested <command> or help <a-command>

        This is called after the following has already happened:
            - docopt has decided argv was valid as given by specs
            - no command was found, or the command did not identify a method
            - if none of the above, the command called parser.help()

        Raises:
            ValueError(message + hint) if no command was found
        """
        command_requested = self.args.get(self.command_tag)
        help_requested = command_requested == 'help' or self.args.get('help')
        action_requested = self.args.get('<any-command>')
        closest_command = None
        # if help was actually requested, see if we can find a command
        if help_requested:
            if not action_requested:
                docs = self.docs
                self.logger.info(dedent(docs).strip())
                return
            command_requested = action_requested
            usage_only = False
        for commandcls in self.commands:
            if commandcls.command == command_requested:
                command = self._command_instance(commandcls)
                return command.help(usage_only=usage_only)
            if command_requested in commandcls.command or commandcls.command in command_requested:
                closest_command = commandcls.command
        hint = f"try: om help {closest_command}" if closest_command else 'try: om -h'
        msg = f"sorry, I don't know about >{command_requested}<. {hint}"
        raise SystemExit(msg)

    def ask(self, prompt, hide=False, options=None, default=None):
        """
        ask user input

        Can be overridden by providing CommandParser(..., askfn=callable), e.g.
        for testing purpose

        Args:
            prompt (str): the prompt text
            hide (bool): if True don't ask, just return default
            options (str): options string, if provided answer will be checked
                to be in options (lowercase)
            default (str): the default answer if no answer given or self.silent
                is active (-q switch)

        Returns:
            answer as str
        """
        if self._askfn:
            return self._askfn(prompt, hide=hide, options=options, default=default,
                               parser=self)
        if not self.silent:
            options_text = f'[{options}] ' if options else ''
            prompt = f'{prompt} {options_text}'
            if hide:
                value = getpass(prompt=prompt)
            value = input(prompt)
            if not value and default:
                return default
            if options:
                assert value.lower() in options.lower()
        return default


class CommandBase:
    """
    Base class for CommandParser commands

    This assumes either of the following docopt formats:

        pgm.py <command> [<action>]
        pgm.py command action

    How to use:
        1. implement a CommandBase subclass for the command you wish to handle
        2. implement methods for each action (e.g. action <dothis> gets method dothis)
        3. to process command without action, implement __call__
        4. Pass your CommandBase classes as the command list to CommandParser

        Each CommandBase subclass must at least provide the command name
        and either a valid action method or the __call__ method.

        Each action method, including __call__, is of format
            def action(self):
                # use self.args to get parsed arguments

        If a method with equal name as the action is found, it will be called.
        If no method is found, __call__ will be invoked, which by default prints
        the command's help

        Note a named <argument> will be used to lookup a method based on the
        value passed. e.g if <argument> is specified as foo then the foo method
        will be found.

        See CommandParser for a complete example.

    Documentation of commands:
        A command class can have its own docstring in the same format as the
        main docstring, with its own Usage: and Options: sections. If it does
        so, docopt is automatically called on the command's docstring
        and self.args is updated accordingly.

        The command's Usage and Options sections can be included in the main
        docstring by adding [usage:<command>] and [options:<command>]
        placeholders. In this case the Usage: and Options: strings will be
        replaced by usage_header and options_header, respectively.

        See CommandParser for a complete example.
    """
    command = 'unspecified'
    action_tag = '<action>'
    usage_header = 'Usage of {self.command}'
    options_header = 'Options for {self.command}'
    description_header = 'Working with {self.command}'

    def __init__(self, docs, argv=None, logger=None, parser=None):
        # initialize, if this class contains a __doc__ header
        # with a Usage: section, self.args will automatically
        # be updated accordingly.
        self.argv = argv or sys.argv[1:]
        self.logger = logger
        self.docs = docs or self.__doc__
        self.parser = parser
        self.global_docs = self.parser.docs if self.parser else None
        self.args = {}

    def parse(self):
        """
        Use docopt on self.docs and self.args

        If this method returns, self.args will contain the parsed options
        as given in self.docs
        """
        if self.docs is None:
            return
        docs = self.add_global_options() if self.global_docs else self.docs
        args = safe_docopt(docs, argv=self.argv, help=False)
        self.args.update(args)
        if self.parser.should_debug:
            print("*** {self.command} parsed args".format(**locals()), self.args)

    @property
    def has_usage(self):
        return self.docs and 'usage:' in self.__doc__.lower()

    @property
    def has_description(self):
        return self.docs and 'description:' in self.__doc__.lower()

    @property
    def usage(self):
        lines = []
        for line in self.docs.split('\n'):
            if any(v in line.lower() for v in ('options:', 'description:')):
                break
            lines.append(line)
        return '\n'.join(lines)

    @property
    def silent(self):
        return self.parser.silent

    def add_global_options(self):
        # add Options: from top level parser. this is to make sure
        # that global options are checked on any [options] placeholder
        # in this command's docstring
        if self.global_docs and 'options:' in self.global_docs.lower():
            # get global options as clean as possible
            global_options = self.global_docs.split('Options:', 1)[1]
            global_options = global_options.split('\n')
            docs = dedent(self.docs).split('\n')
            # find options to add, stopping at the first empty line
            # after options. note the first line is always empty
            to_add = []
            for i, opt in enumerate(global_options):
                if i > 0 and not opt:
                    break
                if opt.strip().split('  ')[0] not in self.docs:
                    to_add.append(opt)
            docs.extend(to_add)

            def section(line):
                if line.lower().startswith('options'):
                    line = '\n' + line
                return line

            docs = '\n'.join(section(line) for line in docs)
        else:
            docs = self.docs
        return docs

    def help(self, usage_only=False):
        """
        Display command's help

        This is called by the parser upon the help command, and if __call__
        is not implemented.

        Args:
            usage_only (bool): if True, show only usage, else show full docs

        Returns:

        """
        if usage_only:
            self.logger.info(dedent(self.usage))
        else:
            self.logger.info(dedent(self.docs))

    def __call__(self):
        """
        If no method was found, this is what gets help. By
        default we call self.help(usage_only=True) to show
        usage information
        """
        self.help(usage_only=True)

    def get_command_method(self):
        """
        get the method to call, by default return command itself

        for <action> style arguments will use the value to lookup.
        The action tag used for the lookup is self.action_tag

        Returns:
              the method or self if none was found. Returning self
              means __call__() will be called
        """
        #
        #
        #
        for k, v in self.args.items():
            lookup = str(v if k == self.action_tag else k)
            if v:
                meth = getattr(self, lookup, None)
                if meth is not None:
                    return meth
        return self

    @classmethod
    def get_command_docparts(self):
        # return usage, options, description to replace
        # [usage:<command>], [options:<command>], [description:<command>]
        # -- format headers
        usage_header = self.usage_header.format(**locals())
        options_header = self.options_header.format(**locals())
        description_header = self.description_header.format(**locals())
        # get clean sections
        clean_command_doc = dedent(self.__doc__)
        clean_command_doc = clean_command_doc.replace('Usage:', 'usage:')
        clean_command_doc = clean_command_doc.replace('Options:', 'options:')
        clean_command_doc = clean_command_doc.replace('Description:', 'description:')
        # ensure all sections are present
        if 'options:' not in clean_command_doc:
            # ensure options is before description
            if 'description:' in clean_command_doc:
                before, after = clean_command_doc.split('description:')
                before += '\noptions:\n'
                clean_command_doc = before + 'description:' + after
            else:
                clean_command_doc += '\noptions:\n'
        if 'description:' not in clean_command_doc:
            clean_command_doc += '\ndescription:\n'
        # split
        try:
            usage, options = clean_command_doc.split('options:')
            options, description = options.split('description:')
        except:
            raise ValueError(f"cannot parse {clean_command_doc}")
        # ensure headers are not interfering with top-level
        usage = usage.replace('usage:', usage_header)
        options = (options_header + options) if options.strip() else ''
        description = (description_header + description) if description.strip() else ''
        return usage, options, description

    def parse_kwargs(self, argname, resolve_bool=True, splitby=None,
                     pyeval=False, literal_escape=None,
                     **defaults):
        """
        Parse [<kw=value>] and [<kw=value>...] kind arguments

        Usage:
            Positional parameters

                prog [<kw=value>]
                prog [<kw=value>...]

            => in your BaseCommand call self.parse_kwargs('<kw=value>') to get back the parsed dict

            Long (named) parameters

                prog [--param=<kw=value>]
                prog [--param=<kw=value>]...

            => in your BaseCommand call self.parse_kwargs('--param') to get back the parsed dict

        Examples:
                a=1    => { a: '1' }
                a=1    => { a: 1 } # assuming pyeval=True
                a=yes  => { a: True }
                a=false => { a: False }
                a=*false => { a: 'false' }
                a=[0,1,2] => { a: [0,1,2] } # assuming pyeval=True
                a=1,2,3   => { a: ['1', '2', '3'] # assuming splitby=', pyeval=False

        Args:
            argname (str): the name of the argument as specified in docopt
            splitby (str): specify string to split values by, by default splits on comma,
                unless pyeval is given. If pyeval is True by default does not split
            resolve_bool (bool): resolve yes/true => True, no/false => False (defaults to True)
            pyeval (bool): eval() the final value, defaults to False, can also be a callable
            literal_escape (str): the escape char to keep yes/true, no/false, defaults to *
                if resolve_bool is True, else no literal escape
            **defaults (dict): any defaults

        Returns:
            dict of kw => value, can contain a single or multiple items
        """
        # replace literal *yes, *no with 'yes', 'no'
        # replace literal *true, *false with 'true', 'false'
        # replace yes/true with True
        # replace no/false with False
        # split values that contain a comma to a list
        splitby = None if pyeval else ','
        literal_escape = '*' if not resolve_bool else None
        values = self.args.get(argname)
        if values is None:
            return {}
        if not isinstance(values, list):
            values = [values]  # allow single values
        # seperate method to make it reusable by parse_kwargs and parse_kwarg
        BOOLMAP = {'true': True, 'yes': True,
                   'no': False, 'false': False,
                   'y': True, 'n': False}
        doeval = lambda v: pyeval(v) if callable(pyeval) else (eval(v) if pyeval else v)
        split = lambda v: ([v for v in v.split(splitby) if v]
                           if splitby and (len(v) > 1 and splitby in v)
                           else v)
        literal = lambda v: v.replace(literal_escape, '') if literal_escape else v
        truefalse = lambda v: BOOLMAP.get(v.lower()) if resolve_bool else v
        resolve = lambda v: truefalse(v) if resolve_bool and v in BOOLMAP else doeval(split(literal(v)))
        requested_kwargs = {k: resolve(v) for arg in values for k, v in [arg.split('=', 1)]}
        real_kwargs = {}
        real_kwargs.update(defaults)
        real_kwargs.update(requested_kwargs)
        return real_kwargs

    def ask(self, *args, **kwargs):
        return self.parser.ask(*args, **kwargs)


def safe_docopt(doc, argv=None, help=True, version=None, options_first=False):
    """
    A docopt drop-in replacement to print meaning full error messages in case
    of a DocoptLanguageError exception

    Args:
        same as docopt

    Returns
        parsed args
    """
    try:
        args = docopt(doc, argv=argv, help=help, version=version, options_first=options_first)
    except DocoptLanguageError as e:
        print("*** ERROR {}, check below".format(e))
        print("arguments to doctopt were")
        print("   argv=", argv)
        print("   doc=", doc)
        print("**** ERROR {}, check above messages ***".format(e))
        exit(1)
    return args


def setup_console_logger():
    """
    Set up a default console logger that prints messages as is

    Returns:
        the logger instance
    """
    logger = logging.getLogger()
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger



from __future__ import absolute_import

import warnings
from importlib import import_module

import logging
import os
import re
import six
import sys
import tempfile
import uuid
from shutil import rmtree
from six import string_types

try:
    import urlparse
except:
    from urllib import parse as urlparse

__settings = None


def is_dataframe(obj):
    try:
        import pandas as pd
        return isinstance(obj, pd.DataFrame)
    except:
        return False


def is_series(obj):
    try:
        import pandas as pd
        return isinstance(obj, pd.Series) and not isinstance(obj, pd.DataFrame)
    except:
        return False


def is_estimator(obj):
    try:
        from sklearn.base import BaseEstimator
        from sklearn.pipeline import Pipeline
        return isinstance(obj, (BaseEstimator, Pipeline))
    except:
        False


def is_ndarray(obj):
    try:
        import numpy as np
        return isinstance(obj, np.ndarray)
    except:
        False


def is_spark_mllib(obj):
    """
    # unlike scikit learn obj is not the actual model, but a specification of
    # the model for the spark server to create. so obj is the name of the
    # python class, e.g. obj=pyspark.mllib.clustering.KMeans
    """
    if isinstance(obj, string_types):
        return 'pyspark.mllib' in obj
    return False


def settings(reload=False):
    """ wrapper to get omega settings from either django or omegamldefaults """
    from omegaml import _base_config as omdefaults
    global __settings
    if not reload and __settings is not None:
        return __settings
    try:
        # see if we're running as a django app
        from django.utils import six
        from django.utils.functional import empty
        from django.conf import settings as djsettings  # @UnresolvedImport
        defaults = djsettings
        # this is to test if django was initialized. if not revert
        # to using omdefaults
        try:
            if defaults._wrapped is empty:
                # django is not initialized, use omega defaults
                raise ValueError()
            getattr(defaults, 'SECRET_KEY')
        except Exception as e:
            from warnings import warn
            warn("Using omegaml.defaults because Django was not initialized."
                 "Try importing omegaml within a method instead of at the "
                 "module level")
            raise
    except Exception as e:
        defaults = omdefaults
    else:
        # get default omega settings into django settings if not set
        # there already
        from omegaml import _base_config as omdefaults
        for k in dir(omdefaults):
            if k.isupper() and not hasattr(defaults, k):
                setattr(defaults, k, getattr(omdefaults, k))
    __settings = defaults
    return __settings


def override_settings(**kwargs):
    """ test support """
    cfgvars = settings()
    for k, v in six.iteritems(kwargs):
        setattr(cfgvars, k, v)
    # -- OMEGA_CELERY_CONFIG updates
    celery_config = getattr(cfgvars, 'OMEGA_CELERY_CONFIG', {})
    for k in [k for k in kwargs.keys() if k.startswith('OMEGA_CELERY')]:
        celery_k = k.replace('OMEGA_', '')
        celery_config[celery_k] = kwargs[k]
    setattr(cfgvars, 'OMEGA_CELERY_CONFIG', celery_config)


def delete_database():
    """ test support """
    from omegaml.mongoshim import MongoClient

    mongo_url = settings().OMEGA_MONGO_URL
    parsed_url = urlparse.urlparse(mongo_url)
    database_name = parsed_url.path[1:]
    # authenticate via admin db
    # see https://stackoverflow.com/a/20554285
    c = MongoClient(mongo_url, authSource='admin')
    c.drop_database(database_name)


def make_tuple(arg):
    if not isinstance(arg, (list, tuple)):
        arg = (arg,)
    return tuple(arg)


def make_list(arg):
    if not isinstance(arg, (list)):
        arg = list(arg)
    return arg


def flatten_columns(col, sep='_'):
    # source http://stackoverflow.com/a/29437514
    if not type(col) is tuple:
        return col
    else:
        new_col = ''
        for leveli, level in enumerate(col):
            if not level == '':
                if not leveli == 0:
                    new_col += sep
                new_col += level
        return new_col


CLASS_CACHE = {}


def load_class(requested_class):
    """
    Check if requested_class is a string, if so attempt to load
    class from module, otherwise return requested_class as is
    """
    import importlib
    if isinstance(requested_class, six.string_types):
        if requested_class in CLASS_CACHE:
            return CLASS_CACHE.get(requested_class)
        module_name, class_name = requested_class.rsplit(".", 1)
        try:
            m = importlib.import_module(module_name)
            cls = getattr(m, class_name)
            CLASS_CACHE[requested_class] = cls
            return cls
        except Exception as e:
            logging.debug(
                'could not load module %s for class %s due to %s. sys.path=%s' % (
                    module_name, class_name, str(e), str(sys.path)))
            raise
    return requested_class


def get_rdd_from_df(df):
    """
    takes a pandas df and returns a spark RDD
    """
    from pyspark import SparkContext, SQLContext
    from pyspark.mllib.linalg import Vectors
    sc = SparkContext.getOrCreate()
    from warnings import warn
    warn(
        "get_rdd_from_df creates a spark context, it is recommended"
        " that you use SparkContext.getOrCreate() to prevent multiple context"
        " creation")
    sqlContext = SQLContext(sc)
    spark_df = sqlContext.createDataFrame(df)
    rdd = spark_df.rdd.map(lambda data: Vectors.dense(
        [float(x) for x in data]))
    return rdd


def get_labeledpoints(Xname, Yname):
    """
    returns a labeledpoint RDD from the datasets provided
    """
    import omegaml as om
    from pyspark.mllib.regression import LabeledPoint
    # import from datastore
    X = om.datasets.get(Xname)
    Y = om.datasets.get(Yname)
    result = Y.join(X)
    # create labeled point
    rdd = get_rdd_from_df(result)
    labeled_point = rdd.map(lambda x: LabeledPoint(float(x[0]), x[1:]))
    return labeled_point


def get_labeled_points_from_rdd(rdd):
    """
    returns a labeledpoint from the RDD provided
    """
    from pyspark.mllib.regression import LabeledPoint
    return rdd.map(lambda x: LabeledPoint(float(x[0]), x[1:]))


def unravel_index(df, row_count=0):
    """
    convert index columns into dataframe columns

    index columns are stored in the dataframe, named '_idx#<n>_<name>'
    where n is the sequence and name is the nth index name.

    Use restore_index_columns_order to get back the original index columns
    in sequence.

    :param df: the dataframe
    :param row_count: the row_count base, ensures subsequent stores
      get differnt row ids
    :return: the unravelled dataframe, meta
    """
    # remember original names
    idx_meta = {
        'names': df.index.names,
    }
    # convert index names so we can restore them later
    store_idxnames = ['_idx#{}_{}'.format(i, name or i)
                      for i, name in enumerate(idx_meta['names'])]
    df.index.names = store_idxnames
    unravelled_df, idx_meta = df.reset_index(), idx_meta
    # store row ids
    unravelled_df['_om#rowid'] = unravelled_df.index.values + row_count
    # restore index names on original dataframe
    df.index.names = idx_meta['names']
    return unravelled_df, idx_meta


def restore_index_columns_order(columns):
    """
    from an iterable of column names get the index columns in sequence order

    index columns are named '_idx#<n>_<name>' where n is the sequence
    of the original index column and name is the name
    """

    def get_index_order(col):
        if '_idx#' in col:
            n = col.split('_')[1].split('#')[1]
        else:
            n = 0
        return n

    index_cols = (col for col in columns if isinstance(col, six.string_types) and col.startswith('_idx'))
    index_cols = sorted(index_cols, key=get_index_order)
    return index_cols


def restore_index(df, idx_meta, rowid_sort=True):
    """
    restore index proper

    :param df: the dataframe
    :param idx_meta: index metadata
    :param rowid_sort: whether to sort by row id. defaults to True
           If your query is already sorted in some specific way,
           specify False to keep the sort order.
    """
    # -- establish row order proper
    if rowid_sort and '_om#rowid' in df:
        df.sort_values('_om#rowid', inplace=True)
        del df['_om#rowid']
    # -- get index columns
    index_cols = restore_index_columns_order(df.columns)
    # -- set index columns
    result = df.set_index(index_cols) if index_cols else df
    if index_cols:
        result.index.names = idx_meta.get('names', [None] * len(index_cols))
    return result


def jsonescape(s):
    return str(s).replace('.', '_')


def grouper(n, iterable):
    # https://stackoverflow.com/a/8998040
    import itertools
    it = iter(iterable)
    while True:
        chunk_it = itertools.islice(it, n)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield itertools.chain((first_el,), chunk_it)


def cursor_to_dataframe(cursor, chunk_size=10000, parser=None):
    # a faster and less memory hungry variant of DataFrame.from_records
    # works by building a set of smaller dataframes to reduce memory
    # consumption. Note chunks are of size max. chunk_size.
    import pandas as pd
    frames = []
    if hasattr(cursor, 'count'):
        count = cursor.count()
        chunk_size = max(chunk_size, int(count * .1))
    else:
        # CommandCursors don't have .count, go as long as we can
        count = None
    if count is None or count > 0:
        for chunk in grouper(chunk_size, cursor):
            df = pd.DataFrame.from_records(chunk) if not parser else parser(r for r in chunk)
            frames.append(df)
        if frames:
            df = pd.concat(frames)
        else:
            df = pd.DataFrame()
    else:
        df = pd.DataFrame()
    return df


def reshaped(data):
    """
    check if data is 1d and if so reshape to a column vector
    """
    import pandas as pd
    import numpy as np
    if isinstance(data, (pd.Series, pd.DataFrame)):
        if len(data.shape) == 1:
            data = data.values.reshape(-1, 1)
        else:
            if len(data.shape) == 2 and data.shape[1] == 1:
                data = data.values.reshape(-1, 1)
    elif isinstance(data, np.ndarray):
        if len(data.shape) == 1:
            data = data.reshape(-1, 1)
    elif isinstance(data, list):
        data = np.array(data)
        if len(data.shape) == 1:
            data = data.reshape(-1, 1)
    return data


def gsreshaped(data):
    """
    gridsearch reshape values according to GridSearchCV.fit

    see https://stackoverflow.com/a/49241326
    """
    import pandas as pd
    import numpy as np
    if isinstance(data, (pd.Series, pd.DataFrame)):
        if len(data.shape) == 2 and data.shape[1] == 1:
            data = data.values.reshape(-1)
    elif isinstance(data, np.ndarray):
        if len(data.shape) == 2 and data.shape[1] == 1:
            data = data.reshape(-1)
    return data


def convert_dtypes(df, dtypes):
    """
    get back original dtypes

    :param df: the dataframe to apply conversion to
    :param dtypes: the dict mapping column to dtype (use kind_meta['dtypes'])
    """
    # tz pattern used in convert_dtypes
    tzinfo_pattern = re.compile('datetime64\[ns, (.*)\]')
    for col, dtype in six.iteritems(dtypes):
        if dtype.startswith('datetime'):
            if not hasattr(df, 'dtypes'):
                continue
            try:
                match = tzinfo_pattern.match(dtype)
                if match:
                    tzname = match.groups()[0]
                    df[col] = df[col].dt.tz_localize('UTC').dt.tz_convert(tzname)
            except:
                # TODO ignore errors, issue warning
                pass

    return df


class PickableCollection(object):
    def __init__(self, collection):
        super(PickableCollection, self).__setattr__('collection', collection)

    def __getattr__(self, k):
        return getattr(self.collection, k)

    def __setattr__(self, k, v):
        return setattr(self.collection, k, v)

    def __getitem__(self, k):
        return self.collection[k]

    def __setitem__(self, k, v):
        self.collection[k] = v

    def __getstate__(self):
        client = self.collection._Collection__database._Database__client
        host, port = list(client.nodes)[0]
        return {
            'name': self.name,
            'database': self.database.name,
            'host': host,
            'port': port,
            'credentials': self.database.client._MongoClient__all_credentials[self.database.name],
        }

    def __setstate__(self, state):
        from omegaml.mongoshim import MongoClient
        url = 'mongodb://{credentials.username}:{credentials.password}@{host}:{port}/{database}'.format(**state)
        client = MongoClient(url, authSource=state['credentials'].source)
        db = client.get_database()
        collection = db[state['name']]
        super(PickableCollection, self).__setattr__('collection', collection)

    def __repr__(self):
        return 'PickableCollection({})'.format(repr(self.collection))


def extend_instance(obj, cls, *args, **kwargs):
    """Apply mixins to a class instance after creation"""
    # source https://stackoverflow.com/a/31075641
    from omegaml import load_class
    cls = load_class(cls)
    if cls not in obj.__class__.mro():
        base_cls = obj.__class__
        base_cls_name = 'Extended{}'.format(obj.__class__.__name__.split('.')[0])
        obj.__class__ = type(base_cls_name, (cls, base_cls), {})
    if hasattr(obj, '_init_mixin'):
        obj._init_mixin(*args, **kwargs)


def temp_filename(dir=None, ext='tmp'):
    """ generate a temporary file name """
    dir = dir or tempfile.mkdtemp()
    return os.path.join(dir, uuid.uuid4().hex + f'.{ext}')


def remove_temp_filename(fn, dir=True):
    dirname = os.path.dirname(fn)
    os.remove(fn)
    if dir:
        if dirname.startswith('/tmp/') and len(dirname.split('/')) > 1:
            rmtree(dirname)
        else:
            warnings.warn('will not remove directory {} as it is outside of /tmp'.format(fn))


def ensure_python_array(arr, dtype):
    import numpy as np
    return np.array(arr).astype(dtype)


def dict_update_if(condition, dict, other):
    try:
        condition()
    except:
        pass
    else:
        dict.update(other)


def module_available(modname):
    try:
        import_module(modname)
    except:
        return False
    return True


def tensorflow_available():
    return module_available('tensorflow')


def keras_available():
    return module_available('keras')


def calltrace(obj):
    """
    trace calls on an object

    Usage:
        .. code ::

            def __init__(self, *args, **kwargs):
                calltrace(self)

        This will print the method arguments and return values
        for every call. Exceptions will be caught and printed,
        then re-raised. Magic methods are not traced.

    Args:
        obj: the object to trace
    """

    def tracefn(el):
        def methodcall(*args, **kwargs):
            print("calling", el, args, kwargs)
            try:
                result = el(*args, **kwargs)
            except (BaseException, Exception) as e:
                print("=> ", str(e))
                raise
            except:
                print("=> unknown exception")
                raise
            else:
                print("=> ", result)
            return result

        return methodcall

    for item in dir(obj):
        if item.startswith('__'):
            continue
        el = getattr(obj, item)
        if type(el).__name__ == 'method':
            setattr(obj, item, tracefn(el))
    return obj


class DefaultsContext(object):
    def __init__(self, source):
        for k in dir(source):
            if k.isupper():
                setattr(self, k, getattr(source, k))

    def __iter__(self):
        for k in dir(self):
            if k.startswith('OMEGA') and k.isupper():
                yield k, getattr(self, k)

    def __repr__(self):
        return '{}'.format(dict(self))


def ensure_json_serializable(v):
    import numpy as np
    import pandas as  pd
    if isinstance(v, np.ndarray):
        return v.flatten().tolist()
    if isinstance(v, pd.Series):
        return ensure_json_serializable(v.to_dict())
    elif isinstance(v, dict):
        vv = {
            k: ensure_json_serializable(v)
            for k, v in six.iteritems(v)
        }
        return vv
    return v


def mkdirs(path):
    """ safe os.makedirs for python 2 & 3
    """
    if not os.path.exists(path):
        os.makedirs(path)



from __future__ import absolute_import

import os
import random
import string
import unittest
from unittest.case import TestCase, skip

import numpy as np
import pandas as pd
from omegaml import Omega
from omegaml.mdataframe import MDataFrame
from omegaml.util import flatten_columns
from pandas.util.testing import assert_frame_equal, assert_series_equal
from six.moves import range

class MDataFrameTests(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        df = self.df = pd.DataFrame({'x': list(range(0, 10)) + list(range(0, 10)),
                                     'y': random.sample(list(range(0, 100)), 20)})
        om = self.om = Omega()
        om.datasets.put(df, 'sample', append=False)
        self.coll = om.datasets.collection('sample')

    def tearDown(self):
        TestCase.tearDown(self)

    def test_groupby(self):
        coll = self.coll
        df = self.df
        keys = []
        for key, groupdf in MDataFrame(coll).groupby(['x']):
            x = key.get('x')
            keys.append(x)
            subdf = df[df.x == x]
            assert_frame_equal(subdf, groupdf.value)
        self.assertEqual(set(keys), set(df.x))

    def test_count(self):
        coll = self.coll
        df = self.df
        counts = MDataFrame(coll).groupby(['x']).count()
        test_counts = df.groupby('x').count()
        self.assertTrue(test_counts.equals(counts))

    def test_count_multi_columns(self):
        coll = self.coll
        df = self.df
        # add a column
        mdf = MDataFrame(coll)
        mdf['z'] = 5
        df['z'] = 5
        # group by and count
        counts = mdf.groupby(['x']).count()
        test_counts = df.groupby('x').count()
        self.assertTrue(test_counts.equals(counts))

    def test_count_column(self):
        coll = self.coll
        df = self.df
        result = MDataFrame(coll).groupby(['x']).x.count()
        testgroup = df.groupby('x').x.count()
        self.assertTrue(result.equals(testgroup))

    def test_aggregate(self):
        coll = self.coll
        df = self.df
        result = MDataFrame(coll).groupby(['x']).agg({'x': 'sum'})
        testagg = df.groupby('x').agg({'x': 'sum'})
        testagg = testagg.rename(columns=dict(x='x_sum'))
        self.assertTrue(result.equals(testagg))

    def test_aggregate_multi_stats(self):
        coll = self.coll
        df = self.df
        stats = {'x': ['sum', 'mean', 'max', 'min', 'std']}
        result = MDataFrame(coll).groupby(['x']).agg(stats)
        testagg = df.groupby('x').agg(stats)
        testagg.columns = testagg.columns.map(flatten_columns)
        testagg = testagg[result.columns]
        assert_frame_equal(testagg, result, check_dtype=False)

    def test_mdataframe(self):
        coll = self.coll
        df = self.df
        mdf = MDataFrame(coll)
        result = mdf.value
        self.assertEqual(set(MDataFrame(coll).columns),
                         set(list(df.columns)))
        self.assertTrue(result.equals(df))
        self.assertEqual(mdf.shape, df.shape)

    def test_mdataframe_count(self):
        coll = self.coll
        df = self.df
        mdf = MDataFrame(coll)
        assert_series_equal(df.count(), mdf.count())
        self.assertEqual(len(mdf), len(mdf))

    def test_mdataframe_xlarge(self):
        df = pd.DataFrame({
            'a': list(range(0, int(1e4 + 1))),
            'b': list(range(0, int(1e4 + 1)))
        })
        store = self.om.datasets
        store.put(df, 'mydata-xlarge', append=False)
        coll = store.collection('mydata-xlarge')
        result = MDataFrame(coll).value
        self.assertEqual(set(MDataFrame(coll).columns),
                         set(list(df.columns)))
        self.assertTrue(result.equals(df))

    def test_mdataframe_column_attribute(self):
        coll = self.coll
        df = self.df
        result = MDataFrame(coll).y.value
        self.assertTrue(df.y.equals(result))

    def test_mdataframe_column_slice(self):
        coll = self.coll
        df = self.df
        result = MDataFrame(coll)['y'].value
        self.assertTrue(df['y'].equals(result))

    def test_mdataframe_columns_slice(self):
        coll = self.coll
        df = self.df
        result = MDataFrame(coll)[['x', 'y']].value
        self.assertTrue(df[['x', 'y']].equals(result))

    def test_mdataframe_sort(self):
        coll = self.coll
        df = self.df
        result = MDataFrame(coll).sort(['-x', '-y']).value
        df = df.sort_values(
            ['x', 'y'], ascending=[False, False])
        assert_frame_equal(df, result)

    def test_mdataframe_merge(self):
        coll = self.coll
        df = self.df
        om = self.om
        other = pd.DataFrame({'x': list(range(0, 20)),
                              'y': list(range(0, 20)),
                              'z': list(range(0, 20))})
        om.datasets.put(other, 'samplez', append=False)
        coll2 = om.datasets.collection('samplez')
        result = MDataFrame(coll).merge(coll2, on='x', how='left').value
        testdf = df.merge(other, on='x', how='left')
        self.assertTrue(result.equals(testdf))

    def test_mdataframe_merge_differing_columns(self):
        coll = self.coll
        df = self.df
        om = self.om
        other = pd.DataFrame({'s': list(range(0, 20)),
                              'y': list(range(0, 20)),
                              'z': list(range(0, 20))})
        om.datasets.put(other, 'samplez', append=False)
        coll2 = om.datasets.collection('samplez')
        result = MDataFrame(coll).merge(coll2, left_on='x',
                                        right_on='s', how='left').value
        testdf = df.merge(other, left_on='x', right_on='s', how='left')
        testdf = testdf[result.columns]
        self.assertTrue(result.equals(testdf))

    def test_mdataframe_merge_nomatch(self):
        coll = self.coll
        df = self.df
        om = self.om
        other = pd.DataFrame({'x': list(range(50, 55)),
                              'y': list(range(0, 5)),
                              'z': list(range(0, 5))})
        om.datasets.put(other, 'samplez', append=False)
        coll2 = om.datasets.collection('samplez')
        result = MDataFrame(coll).merge(coll2, on='x', how='left').value
        testdf = df.merge(other, on='x', how='left')
        testdf = testdf[result.columns]
        self.assertTrue(result.equals(testdf))

    @unittest.skip("disalbed because MDataFrame.append fails and is very slow")
    def test_mdataframe_merge_append(self):
        ## FIXME this does not work
        coll = self.coll
        df = self.df
        om = self.om
        other = pd.DataFrame({'x': list(range(0, 5)),
                              'y': list(range(0, 5)),
                              'z': list(range(0, 5))})
        om.datasets.put(other, 'samplez', append=False)
        mdf = om.datasets.getl('samplez')
        mdf.append(mdf)
        coll2 = om.datasets.collection('samplez')
        result = MDataFrame(coll).merge(coll2, on='x', how='left', suffixes=('','')).value
        testdf = df.append(other, ignore_index=True)
        testdf = testdf[result.columns]
        assert_frame_equal(result, testdf)

    def test_mdataframe_merge_partial_match(self):
        coll = self.coll
        df = self.df
        om = self.om
        other = pd.DataFrame({'x': list(range(0, 5)),
                              'y': list(range(0, 5)),
                              'z': list(range(0, 5))})
        om.datasets.put(other, 'samplez', append=False)
        coll2 = om.datasets.collection('samplez')
        result = MDataFrame(coll).merge(coll2, on='x', how='left').value
        testdf = df.merge(other, on='x', how='left')
        testdf = testdf[result.columns]
        self.assertTrue(result.equals(testdf))

    def test_mdataframe_merge_inner(self):
        coll = self.coll
        df = self.df
        om = self.om
        other = pd.DataFrame({'x': list(range(0, 5)),
                              'y': list(range(0, 5)),
                              'z': list(range(0, 5))})
        om.datasets.put(other, 'samplez', append=False)
        coll2 = om.datasets.collection('samplez')
        result = MDataFrame(coll).merge(coll2, on='x', how='inner',
                                        sort=True).value
        testdf = df.merge(other, on='x', how='inner', sort=True)
        testdf = testdf[result.columns]
        self.assertTrue(result.equals(testdf))

    def test_mdataframe_merge_right(self):
        coll = self.coll
        df = self.df
        om = self.om
        other = pd.DataFrame({'x': list(range(0, 5)),
                              'y': list(range(0, 5)),
                              'z': list(range(0, 5))})
        om.datasets.put(other, 'samplez', append=False)
        coll2 = om.datasets.collection('samplez')
        result = MDataFrame(coll).merge(coll2, on='x', how='left',
                                        sort=True).value
        testdf = df.merge(other, on='x', how='left', sort=True)
        testdf = testdf[result.columns]
        self.assertTrue(result.equals(testdf))

    def test_verylarge_dataframe(self):
        if not os.environ.get('TEST_LARGE'):
            return
        other = pd.DataFrame({'x': list(range(0, int(10e6))),
                              'y': list(range(0, int(10e6))),
                              'z': list(range(0, int(10e6)))})
        coll = self.coll
        df = self.df
        result = MDataFrame(coll).value
        self.assertEqual(set(MDataFrame(coll).columns),
                         set(list(df.columns)))
        self.assertTrue(result.equals(df))

    def test_unique_series(self):
        coll = self.coll
        df = self.df
        om = self.om
        om.datasets.put(df, 'uniques', append=False)
        coll = om.datasets.collection('uniques')
        result = MDataFrame(coll).x.unique().value
        self.assertListEqual(list(result), list(df.x.unique()))

    def test_query_null(self):
        om = self.om
        df = pd.DataFrame({'x': list(range(0, 5)),
                           'y': [1, 2, 3, None, None]})
        om.datasets.put(df, 'foox', append=False)
        result = om.datasets.get('foox', y__isnull=True, lazy=True).value
        test = df[df.isnull().any(axis=1)]
        assert_frame_equal(result, test)

    def test_locindexer_numeric_index(self):
        om = self.om
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        df = pd.DataFrame(data)
        om.datasets.put(df, 'foo', append=False)
        # by label
        dfx = om.datasets.getl('foo').loc[4].value
        assert_series_equal(df.loc[4], dfx)
        # by slice
        dfx = om.datasets.getl('foo').loc[2:4].value
        assert_frame_equal(df.loc[2:4], dfx)
        # by list
        dfx = om.datasets.getl('foo').loc[[2, 4]].value
        assert_frame_equal(df.loc[[2, 4]], dfx)
        # by ndarray
        sel = np.array([1, 2])
        dfx = om.datasets.getl('foo').loc[sel, :].value
        assert_frame_equal(df.loc[sel, :], dfx, check_names=False)

    def test_locindexer_character_index(self):
        om = self.om
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        idx = string.ascii_lowercase[0:9]
        df = pd.DataFrame(data, index=(c for c in idx))
        om.datasets.put(df, 'foo', append=False)
        # by label
        dfx = om.datasets.getl('foo').loc['c'].value
        assert_series_equal(df.loc['c'], dfx)
        # by slice
        dfx = om.datasets.getl('foo').loc['c':'f'].value
        assert_frame_equal(df.loc['c':'f'], dfx)
        # by list
        dfx = om.datasets.getl('foo').loc[['c', 'f']].value
        assert_frame_equal(df.loc[['c', 'f']], dfx)

    def test_locindexer_timeseries_index(self):
        om = self.om
        # create some dataframe
        tsidx = pd.date_range(pd.datetime(2016, 1, 1), pd.datetime(2016, 4, 1))
        df = pd.DataFrame({
            'a': list(range(0, len(tsidx))),
            'b': list(range(0, len(tsidx)))
        }, index=tsidx)
        om.datasets.put(df, 'foo', append=False)
        # by label
        dfx = om.datasets.getl('foo').loc[pd.datetime(2016, 2, 3)].value
        assert_series_equal(dfx, df.loc[pd.datetime(2016, 2, 3)])
        # by slice
        start, end = pd.datetime(2016, 2, 3), pd.datetime(2016, 2, 8)
        dfx = om.datasets.getl('foo').loc[start:end].value
        assert_frame_equal(df.loc[start:end], dfx)

    def test_locindexer_multiindex(self):
        # create some dataframe
        om = self.om
        midx = pd.MultiIndex(levels=[[u'bar', u'baz', u'foo', u'qux'],
                                     [u'one', u'two']],
                             codes=[
                                 [0, 0, 1, 1, 2, 2, 3, 3],
                                 [0, 1, 0, 1, 0, 1, 0, 1]],
                             names=[u'first', u'second'])
        df = pd.DataFrame({'x': range(0, len(midx))}, index=midx)
        om.datasets.put(df, 'foomidx', append=False)
        dfx = om.datasets.getl('foomidx').loc['bar', 'one'].value
        assert_series_equal(dfx, df.loc['bar', 'one'])

    def test_locindexer_series(self):
        """ test storing a pandas series with it's own index """
        om = self.om
        series = pd.Series(range(10),
                           name='foo',
                           index=pd.date_range(pd.datetime(2016, 1, 1),
                                               pd.datetime(2016, 1, 10)))
        om.datasets.put(series, 'fooseries', append=False)
        # try data range
        daterange = slice(pd.datetime(2016, 1, 5), pd.datetime(2016, 1, 10))
        series2 = om.datasets.getl('fooseries').loc[daterange].value
        assert_series_equal(series2, series.loc[daterange])
        # try single date
        daterange = pd.datetime(2016, 1, 5)
        series2 = om.datasets.getl('fooseries').loc[daterange].value
        self.assertEqual(series2, series.loc[daterange])

    def test_ilocindexer(self):
        om = self.om
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        idx = string.ascii_lowercase[0:9]
        df = pd.DataFrame(data, index=(c for c in idx))
        om.datasets.put(df, 'foo', append=False)
        # by single location
        dfx = om.datasets.getl('foo').iloc[0].value
        assert_series_equal(df.iloc[0], dfx)
        # by slice
        dfx = om.datasets.getl('foo').iloc[0:1].value
        assert_frame_equal(df.iloc[0:1], dfx)
        # by list
        dfx = om.datasets.getl('foo').iloc[[1, 2]].value
        assert_frame_equal(df.iloc[[1, 2]], dfx)


    def test_ilocindexer_single_column(self):
        om = self.om
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        idx = string.ascii_lowercase[0:9]
        df = pd.DataFrame(data, index=(c for c in idx))
        om.datasets.put(df, 'foo', append=False)
        # by single location
        dfx = om.datasets.getl('foo').iloc[0, 1].value
        self.assertEqual(df.iloc[0, 1], dfx)
        # by slice
        # FIXME column access by iloc is not guaranteed to return in order
        dfx = om.datasets.getl('foo').iloc[0:2, 1].value
        assert_series_equal(df.iloc[0:2, 1], dfx, check_names=False)
        # by list
        dfx = om.datasets.getl('foo').iloc[[1, 2], 1].value
        assert_series_equal(df.iloc[[1, 2], 1], dfx, check_names=False)

    def test_ilocindexer_columns(self):
        om = self.om
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        idx = string.ascii_lowercase[0:9]
        df = pd.DataFrame(data, index=(c for c in idx))
        om.datasets.put(df, 'foo', append=False)
        # by single location
        dfx = om.datasets.getl('foo').iloc[0, :].value
        assert_series_equal(df.iloc[0, :], dfx)
        # by slice
        # FIXME column access by iloc is not guaranteed to return in order
        dfx = om.datasets.getl('foo').iloc[0:2, :].value
        assert_frame_equal(df.iloc[0:2, :], dfx, check_names=False)
        # by list
        dfx = om.datasets.getl('foo').iloc[[1, 2], :].value
        assert_frame_equal(df.iloc[[1, 2], :], dfx, check_names=False)
        # by ndarray
        sel = np.array([1,2])
        dfx = om.datasets.getl('foo').iloc[sel, :].value
        assert_frame_equal(df.iloc[sel, :], dfx, check_names=False)

    def test_ilocindexer_array(self):
        om = self.om
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        idx = string.ascii_lowercase[0:9]
        df = pd.DataFrame(data, index=(c for c in idx))
        om.datasets.put(df, 'foo', append=False)
        # by ndarray with immediate loc
        sel = np.array([1,2])
        dfx = om.datasets.getl('foo')
        dfx.immediate_loc = True
        dfx = dfx[['a']].iloc[sel]
        assert_frame_equal(df[['a']].iloc[sel], dfx, check_names=False)
        # by ndarray with delayed loc
        sel = np.array([1, 2])
        dfx = om.datasets.getl('foo')
        dfx.immediate_loc = False
        dfx = dfx[['a']].iloc[sel].value
        assert_frame_equal(df[['a']].iloc[sel], dfx, check_names=False)





from __future__ import absolute_import

import datetime
from importlib import import_module
import os

from omegaml.runtimes.jobproxy import OmegaJobProxy


class DaskTask(object):
    """
    A dask remote function wrapper mimicking a Celery task
    """

    def __init__(self, fn, client, pure=True, **kwargs):
        """
        :param fn: (function) the function to be called
        :param client: (dask client) the dask client to use
        :param pure: (bool) whether this is a dask pure function (will
            be cached or not). Defaults to True.
        """
        self.client = client
        self.fn = fn
        self.pure = pure
        self.kwargs = kwargs

    def delay(self, *args, **kwargs):
        """
        submit the function and execute on cluster.  
        """
        kwargs['pure'] = kwargs.get('pure', self.pure)
        if self.kwargs:
            kwargs.update(self.kwargs)
        return DaskAsyncResult(self.client.submit(self.fn, *args, **kwargs))


class DaskAsyncResult(object):
    """
    A dask Future wrapper mimicking a Celery AsyncResult
    """

    def __init__(self, future):
        self.future = future

    def get(self):
        import dask

        if os.environ.get('DASK_DEBUG'):
            with dask.set_options(get=dask.threaded.get):
                return self.future.result()
        return self.future.result()


def daskhello(*args, **kwargs):
    # test function for dask distributed
    return "hello from {} at {}".format(os.getpid(), datetime.datetime.now())


class OmegaRuntimeDask(object):
    """
    omegaml compute cluster gateway to a dask distributed cluster

    set environ DASK_DEBUG=1 to run dask tasks locally
    """

    def __init__(self, omega, dask_url=None):
        self.dask_url = dask_url
        self.omega = omega
        self._client = None

    @property
    def client(self):
        from distributed import Client, LocalCluster
        if self._client is None:
            if os.environ.get('DASK_DEBUG'):
                # http://dask.pydata.org/en/latest/setup/single-distributed.html?highlight=single-threaded#localcluster
                single_threaded = LocalCluster(processes=False)
                self._client = Client(single_threaded)
            else:
                self._client = Client(self.dask_url)
        return self._client

    def model(self, modelname):
        """
        return a model for remote execution
        """
        from omegaml.runtimes.modelproxy import OmegaModelProxy
        return OmegaModelProxy(modelname, runtime=self)

    def job(self, jobname):
        """
        return a job for remote exeuction
        """
        return OmegaJobProxy(jobname, runtime=self)

    def task(self, name, **kwargs):
        """
        retrieve the task function from the task module

        This retrieves the task function and wraps it into a 
        DaskTask. DaskTask mimicks a celery task and is 
        called on the cluster using .delay(), the same way we
        call a celery task. .delay() will return a DaskAsyncResult,
        supporting the celery .get() semantics. This way we can use
        the same proxy objects, as all they do is call .delay() and
        return an AsyncResult. 
        """
        modname, funcname = name.rsplit('.', 1)
        mod = import_module(modname)
        func = getattr(mod, funcname)
        # we pass pure=False to force dask to reevaluate the task
        # http://distributed.readthedocs.io/en/latest/client.html?highlight=pure#pure-functions-by-default
        return DaskTask(func, self.client, pure=False, **kwargs)

    def settings(self):
        """
        return the runtimes's cluster settings
        """
        return self.task('omegaml.tasks.omega_settings').delay().get()

    def ping(self):
        return DaskTask(daskhello, self.client, pure=False)



import re
import pandas as pd
import six


class ProjectedMixin(object):

    """
    A OmegaStore mixin to process column specifications in dataset name
    """
    colspec_pattern = re.compile(r"(?P<name>.*)\[(?P<colspec>.*)\]$")

    def metadata(self, name, *args, **kwargs):
        if isinstance(name, six.string_types):
            name = name.split('[')[0]
        return super(ProjectedMixin, self).metadata(name, *args, **kwargs)

    def get(self, name, *args, **kwargs):
        """
        Return a projected dataset given a name of form name[colspec]

        colspec can be any of

        * a comma separated list of columns, e.g. foo[a,b]
        * an open-ended slice, e.g. foo[a:] => all columns following a, inclusive
        * an closed slice, e.g. foo[a:b] => all columns between a,b, inclusive
        * a close-ended slice, e.g. foo[:b] => all columns up to b, inclusive
        * an empty slice, e.g. foo[:] => all columns
        * a list of columns to exclude, e.g. foo[^b] => all columns except b

        :param name: (str) the name of the dataset, optionally including a
           column specification
        :return: the dataset with projected columns 
        """
        # split base name from specs, get metadata
        if '[' not in name:
            # no column spec in name, avoid projection
            data = super(ProjectedMixin, self).get(name, *args, **kwargs)
        else:
            # column specs in name, get projected data
            data = self._get_data_from_projection(name, *args, **kwargs)
        return data

    def _get_data_from_projection(self, name, *args, **kwargs):
        # see if we can get columns from metadata
        # if so we can specify the columns before getting the data
        specs, name = name, name.split('[', 1)[0]
        meta = self.metadata(name)
        if 'columns' in meta.kind_meta:
            colmap = meta.kind_meta['columns']
            if isinstance(colmap, dict):
                all_columns = list(colmap.keys())[1]
            else:
                # colmap is list of tuples (colname, storedname)
                all_columns = list(zip(*colmap))[1]
            name, columns = columnset(specs, all_columns)
            kwargs['columns'] = columns
            data = super(ProjectedMixin, self).get(name, *args, **kwargs)
        else:
            # we don't have columns in metadata, get the data first
            # only subset on dataframes
            data = super(ProjectedMixin, self).get(name, *args, **kwargs)
            if isinstance(data, pd.DataFrame):
                all_columns = data.columns
                name, columns = columnset(specs, all_columns)
                data = data[columns]
        return data


def columnset(specs, all_columns):
    """
    find the specified columns in data[colspec] in list of all columns

    colspec can be any of

    * a comma separated list of columns, e.g. foo[a,b]
    * an open-ended slice, e.g. foo[a:] => all columns following a, inclusive
    * an closed slice, e.g. foo[a:b] => all columns between a,b, inclusive
    * a close-ended slice, e.g. foo[:b] => all columns up to b, inclusive
    * an empty slice, e.g. foo[:] => all columns
    * a list of columns to exclude, e.g. foo[^b] => all columns except b
    """
    colspec_pattern = re.compile(r"(?P<name>.*)\[(?P<colspec>.*)\]$")
    match = colspec_pattern.match(specs)
    if match is not None:
        specs, colspec = match.groups()
        if ':' in colspec:
            from_col, to_col = colspec.split(':')
            from_i = (all_columns.index(from_col)
                      if from_col else 0)
            to_i = (all_columns.index(to_col)
                    if to_col else len(all_columns)) + 1
            columns = all_columns[from_i:to_i]
        elif colspec.startswith('^'):
            columns = [col for col in all_columns
                       if col not in colspec[1:].split(',')]
        else:
            columns = colspec.split(',')
    else:
        columns = all_columns
    return specs, columns



import datetime
import flask
import numpy as np
import pandas as pd
import six
from flask import request
from flask_restplus import Resource, fields
from mongoengine import DoesNotExist

import omegaml as om
from omegaml.backends.restapi.model import GenericModelResource
from omegaml.restapi.util import strict
from .app import api

isTrue = lambda v: v if isinstance(v, bool) else (
        v.lower() in ['yes', 'y', 't', 'true', '1'])

PredictInput = strict(api).model('ModelInputSchema', {
    'columns': fields.List(fields.String),
    'data': fields.List(fields.Raw),
    'shape': fields.List(fields.Integer),
})

PredictOutput = api.model('PredictOutput', {
    'model': fields.String,
    'result': fields.Raw,
})

DatasetInput = api.model('DatasetInput', {
    'data': fields.Raw,
    'dtypes': fields.Raw,
    'append': fields.Boolean,
})

DatasetIndex = api.model('DatasetIndex', {
    'values': fields.List(fields.Raw),
    'type': fields.String,
})

DatasetQueryOutput = api.model('DatasetQueryOutput', {
    'data': fields.Raw,
    'index': fields.Nested(DatasetIndex)
})


class OmegaResourceMixin(object):
    """
    helper mixin to resolve the request to a configured Omega instance
    """

    @property
    def _omega(self):
        bucket = flask.request.headers.get('bucket')
        return om.setup()[bucket]

    @property
    def _generic_model_resource(self):
        return GenericModelResource(self._omega)


@api.route('/api/v1/ping')
class PingResource(Resource):
    def get(self):
        dt = datetime.datetime.now()
        return {'ping': dt.isoformat()}


@api.route('/api/v1/model/<string:model_id>/predict')
class ModelResource(OmegaResourceMixin, Resource):
    @api.expect(PredictInput, validate=False)
    @api.marshal_with(PredictOutput)
    def put(self, model_id):
        query = flask.request.args
        payload = api.payload
        return self._generic_model_resource.put(model_id, query, payload)


@api.route('/api/v1/dataset/<string:dataset_id>')
class DatasetResource(OmegaResourceMixin, Resource):
    def _restore_filter(self, om, fltparams, name):
        """
        restore filter kwargs for query in om.datasets.get
        """
        # -- get filters as specified on request query args
        fltkwargs = {k: v for k, v in six.iteritems(fltparams)
                     if k not in ['orient', 'limit', 'skip', 'page']}
        # -- get dtypes of dataframe and convert filter values
        metadata = om.datasets.metadata(name)
        kind_meta = metadata.kind_meta or {}
        dtypes = kind_meta.get('dtypes')
        # get numpy/python typemap. this is required for Py3 support
        # adopted from https://stackoverflow.com/a/34919415
        np_typemap = {v: getattr(six.moves.builtins, k)
                      for k, v in np.typeDict.items()
                      if k in vars(six.moves.builtins)}
        for k, v in six.iteritems(fltkwargs):
            # -- get column name without operator (e.g. x__gt => x)
            col = k.split('__')[0]
            value = six.moves.urllib.parse.unquote(v)
            dtype = dtypes.get(str(col))
            if dtype:
                # -- get dtyped value and convert to python type
                value = np_typemap.get(getattr(np, dtype, str), str)(value)
                fltkwargs[k] = value
        return fltkwargs

    @api.marshal_with(DatasetQueryOutput)
    def get(self, dataset_id):
        om = self._omega
        orient = request.args.get('orient', 'dict')
        fltkwargs = self._restore_filter(om, request.args, dataset_id)
        df = om.datasets.getl(dataset_id, filter=fltkwargs).value
        # get index values as python types to support Py3
        index_values = list(df.index.astype('O').values)
        index_type = type(df.index).__name__
        # get data set values as python types to support Py3
        df = df.reset_index(drop=True)
        df.index = df.index.astype('O')
        # convert nan to None
        # https://stackoverflow.com/a/34467382
        data = df.where(pd.notnull(df), None).astype('O').to_dict(orient)
        return {
            'data': data,
            'index': {
                'values': index_values,
                'type': index_type,
            }
        }

    @api.expect(DatasetInput, validate=True)
    @api.response(200, 'updated')
    def put(self, dataset_id):
        om = self._omega
        orient = request.args.get('orient', 'dict')
        if 'append' in request.args:
            append = isTrue(request.args.get('append', 'true'))
        else:
            append = isTrue(api.payload.get('append', 'true'))
        dtypes = api.payload.get('dtypes')
        if orient == 'dict':
            orient = 'columns'
        if dtypes:
            # due to https://github.com/pandas-dev/pandas/issues/14655#issuecomment-260736368
            dtypes = {k: np.dtype(v) for k, v in six.iteritems(dtypes)}
        df = pd.DataFrame.from_dict(api.payload.get('data'),
                                    orient=orient).astype(dtypes)
        om.datasets.put(df, dataset_id, append=append)
        return '', 200

    @api.response(200, 'updated')
    @api.response(404, 'does not exist')
    def delete(self, dataset_id):
        om = self._omega
        try:
            om.datasets.drop(dataset_id)
        except DoesNotExist:
            status = 404
        else:
            status = 200
        return '', status



import glob
import os
from setuptools import setup, find_packages

README = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()
version = open(os.path.join(os.path.dirname(__file__), 'omegaml', 'VERSION')).read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

# extras
hdf_deps = ['tables>=3.2.2']
tf_deps = ['tensorflow==1.15.0']
keras_deps = ['keras==2.2.4']
graph_deps = ['matplotlib==3.1.0', 'seaborn==0.9.0', 'imageio==2.6.1']
dashserve_deps = ['dashserve']
sql_deps = ['sqlalchemy', 'ipython-sql']

setup(
    name='omegaml',
    version=version,
    packages=find_packages(),
    include_package_data=True,
    data_files=[
        ('omegaml/docs/', glob.glob('./docs/source/nb/*.ipynb'))
    ],
    license='Apache 2.0',
    description='An open source DataOps, MLOps platform for humans',
    long_description=README,
    long_description_content_type='text/x-rst',
    url='https://omegaml.io/',
    author='Patrick Senti',
    author_email='patrick.senti@omegaml.io',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Operating System :: POSIX :: Linux',
        'License :: OSI Approved :: Apache Software License',
    ],
    install_requires=[
        'celery==4.2.1',
        'joblib>=0.9.4',
        'jupyter-client>=4.1.1',
        'pymongo>=3.2.2',
        'mongoengine>=0.18.2,<0.19',
        'pandas>=0.17.1',
        'numpy>=1.16.4,<1.18',
        'scipy>=0.17.0',
        'scikit-learn>=0.20,<0.22',
        'PyYAML>=3.11',
        'flask-restplus>=0.12.1',
        'werkzeug~=0.16.1',  # https://github.com/noirbizarre/flask-restplus/issues/777#issuecomment-584365577
        'six>=1.11.0',
        'croniter>=0.3.30',
        'nbformat>=4.0.1',
        'nbconvert>=5.4.1',
        'dill>=0.2.9',
        'tee>=0.0.3',
        'callable-pip>=1.0.0',
        'appdirs>=1.4.3',
        'cron-descriptor>=1.2.24',
        'docopt>=0.6.2',
        'requests>=2.20.0',
        # fix tensorflow pulling wrong version of absl-py,
        # https://github.com/tensorflow/tensorflow/issues/26691#issuecomment-525519742
        'absl-py>=0.8.1',
        'tqdm>=4.32.2',
    ],
    extras_require={
        'graph': graph_deps,
        'hdf': hdf_deps,
        'tensorflow': tf_deps,
        'keras': keras_deps,
        'dashserve': dashserve_deps,
        'sql': sql_deps,
        'all': hdf_deps + tf_deps + keras_deps + graph_deps + dashserve_deps + sql_deps,
    },
    entry_points={
        'console_scripts': ['om=omegaml.client.cli:climain'],
    }
)




from omegaml.client.docoptparser import CommandBase
from omegaml.client.util import get_omega


class ShellCommandBase(CommandBase):
    """
    Usage:
        om shell [options]
    """
    command = 'shell'

    def shell(self):
        om = get_omega(self.args)
        use_ipython = False
        try:
            import IPython
        except:
            self.logger.warn("you should pip install ipython for convenience")
        else:
            use_ipython = True
        # ipython
        if use_ipython:
            IPython.embed(header='omegaml is available as the om variable', colors='neutral')
            return
        # default console
        import code
        try:
            import gnureadline
        except:
            self.logger.warn("you should pip install gnureadline for convenience")
        variables = {}
        variables.update(locals())
        shell = code.InteractiveConsole(locals=variables)
        shell.interact()



import dill
import six
from mongoengine import GridFSProxy

from omegaml.backends.basedata import BaseDataBackend


class VirtualObjectBackend(BaseDataBackend):
    """
    Support arbitrary functions as object handlers

    Virtual object functions can be any callable that provides a __omega_virtual__
    attribute. The callable must support the following signature:

    .. code::

        @virtualobj
        def virtualobjfn(data=None, method='get|put|drop',
                         meta=None, **kwargs):
            ...
            return data

    Note that there is a distinction between storing the function as a virtual object,
    and passing data in or getting data out of the store. It is the responsibility
    of the function to implement the appropriate code for get, put, drop, as well as
    to keep track of the data it actually stores.

    As a convenience virtual object handlers can be implemented as a subclass of
    VirtualObjectHandler

    Usage:
        # create the 'foo' virtual object
        om.datasets.put(virtualobjfn, 'foo')

        # get data from the virtualobj
        om.datasets.get('foo')
        => will call virtualobjfn(method='get')

        # put data into the virtualobj
        om.datasets.put(data, 'foo')
        => will call virtualobjfn(data=data, method='put')

        # drop the virtualfn
        om.datasets.drop('name')
        => will call virtualobjfn(method='drop') and then
           drop the virtual object completely from the storage

    WARNING:

        Virtual objects are executed in the address space of the client or
        runtime context. Make sure that the source of the code is trustworthy.
        Note that this is different from Backends and Mixins as these are
        pro-actively enabled by the administrator of the client or runtime
        context, respectively - virtual objects can be injected by anyone
        who are authorized to write data.
    """
    KIND = 'virtualobj.dill'

    @classmethod
    def supports(self, obj, name, **kwargs):
        return callable(obj) and getattr(obj, '_omega_virtual', False)

    def put(self, obj, name, attributes=None, **kwargs):
        # TODO add obj signing so that only trustworthy sources can put functions
        # ensure we have a dill'able object
        # -- only instances can be dill'ed
        if isinstance(obj, six.class_types):
            obj = obj()
        data = dill.dumps(obj)
        filename = self.model_store._get_obj_store_key(name, '.dill')
        fileid = self.model_store.fs.put(data, filename=filename)
        gridfile = GridFSProxy(grid_id=fileid,
                               db_alias='omega',
                               collection_name=self.model_store.bucket)
        return self.model_store._make_metadata(
            name=name,
            prefix=self.model_store.prefix,
            bucket=self.model_store.bucket,
            kind=self.KIND,
            attributes=attributes,
            gridfile=gridfile).save()

    def get(self, name, version=-1, force_python=False, lazy=False, **kwargs):
        filename = self.model_store._get_obj_store_key(name, '.dill')
        outf = self.model_store.fs.get_version(filename, version=version)
        obj = dill.load(outf)
        return obj

    def predict(self, modelname, xName, rName, **kwargs):
        # make this work as a model backend too
        meta = self.model_store.metadata(modelname)
        handler = self.get(modelname)
        X = self.data_store.get(xName)
        return handler(method='predict', data=X, meta=meta, store=self.model_store)

def virtualobj(fn):
    """
    function decorator to create a virtual object handler from any
    callable

    Args:
        fn: the virtual handler function

    Usage:

        .. code::

            @virtualobj
            def myvirtualobj(data=None, method=None, meta=None, store=None, **kwargs):
                ...

    See:
        VirtualObjectBackend for details on virtual object handlers

    Returns:
        fn
    """
    setattr(fn, '_omega_virtual', True)
    return fn

class VirtualObjectHandler(object):
    """
    Object-oriented API for virtual object functions
    """
    _omega_virtual = True

    def get(self, data=None, meta=None, store=None, **kwargs):
        raise NotImplementedError

    def put(self, data=None, meta=None, store=None, **kwargs):
        raise NotImplementedError

    def drop(self, data=None, meta=None, store=None, **kwargs):
        raise NotImplementedError

    def predict(self, data=None, meta=None, store=None, **kwargs):
        raise NotImplementedError

    def __call__(self, data=None, method=None, meta=None, store=None, **kwargs):
        MAP = {
            'drop': self.drop,
            'get': self.get,
            'put': self.put,
            'predict': self.predict,
        }
        methodfn = MAP[method]
        return methodfn(data=data, meta=meta, store=store, **kwargs)


from unittest import TestCase

import six
from mock import patch

from omegaml.defaults import update_from_config, update_from_obj, update_from_dict


class BareObj(object):
    pass

class ConfigurationTests(TestCase):
    def setUp(self):
        self.defaults = {
            'OMEGA_MONGO_URL': 'foo'
        }

    def test_update_from_configfile(self):
        """
        Test updates from config file works
        """
        data = """
        OMEGA_MONGO_URL: updated-foo
        NOTOMEGA_VALUE: some other value
        """
        cfgfile = six.StringIO(data)
        cfgfile.seek(0)
        update_from_config(self.defaults, cfgfile)
        self.assertIn('OMEGA_MONGO_URL', self.defaults)
        self.assertEqual(self.defaults['OMEGA_MONGO_URL'], 'updated-foo')
        self.assertNotIn('NOTOMEGA_VALUE', self.defaults)

    def test_update_from_obj(self):
        data = BareObj()
        data.OMEGA_MONGO_URL = 'updated-foo'
        update_from_obj(data, self.defaults)
        self.assertIn('OMEGA_MONGO_URL', self.defaults)
        self.assertEqual(self.defaults['OMEGA_MONGO_URL'], 'updated-foo')
        self.assertNotIn('NOTOMEGA_VALUE', self.defaults)

    def test_update_from_dict(self):
        data = {
            'OMEGA_MONGO_URL': 'updated-foo'
        }
        update_from_dict(data, self.defaults)
        self.assertIn('OMEGA_MONGO_URL', self.defaults)
        self.assertEqual(self.defaults['OMEGA_MONGO_URL'], 'updated-foo')
        self.assertNotIn('NOTOMEGA_VALUE', self.defaults)

    def test_config_from_apikey(self):
        """
        Test an Omega instance can be created from user specific configs
        """
        import omegaml as om
        from omegaml.util import settings
        # check we get default without patching
        defaults = settings()
        setup = om.setup
        with patch.object(defaults, 'OMEGA_MONGO_URL') as mock:
            defaults.OMEGA_MONGO_URL = 'foo'
            om = om.setup()
            self.assertEqual(om.datasets.mongo_url, 'foo')
        # reset om.datasets to restored defaults
        om = setup()
        self.assertNotEqual(om.datasets.mongo_url, 'foo')
        # now test we can change the default through config
        # we patch the actual api call to avoid having to set up the user db
        # the objective here is to test get_omega_from_apikey
        with patch('omegaml.client.userconf.get_user_config_from_api') as mock:
            mock.return_value = {
                'objects': [
                   {
                        'data': {
                            'OMEGA_MONGO_URL': 'updated-foo'
                        }
                    }
                ]
            }
            with patch.object(defaults, 'OMEGA_MONGO_URL') as mock:
                from omegaml.client.userconf import get_omega_from_apikey
                defaults.OMEGA_MONGO_URL = 'foo'
                om = setup()
                self.assertEqual(om.datasets.mongo_url, 'foo')
                om = get_omega_from_apikey('foo', 'bar')
                self.assertEqual(om.datasets.mongo_url, 'updated-foo')
        om = setup()
        self.assertNotEqual(om.datasets.mongo_url, 'foo')




from functools import partial

from omegaml.store import Filter
from omegaml.store import MongoQ
from omegaml.store.filtered import FilteredCollection


class FilterOpsMixin(object):
    """
    filter operators on MSeries
    """

    def __getfltop(op):
        def inner(self, other):
            return self.__fltop__(op, other)

        return inner

    def __fltop__(self, op, other):
        # the actual filter operator
        q = None
        for col in self.columns:
            queryop = '{col}__{op}'.format(col=col, op=op)
            qq = MongoQ(**{queryop: other})
            if q is None:
                q = qq
            else:
                q = q | qq
        return Filter(self.collection, q)

    __eq__ = __getfltop('eq')
    __ne__ = __getfltop('ne')
    __lt__ = __getfltop('lt')
    __le__ = __getfltop('lte')
    __gt__ = __getfltop('gt')
    __ge__ = __getfltop('gte')



from omegaml.client.docoptparser import CommandBase


class CatchallCommandBase(CommandBase):
    command = "catchall"

    def __call__(self):
        if self.args.get('--copyright'):
            logger = self.logger
            logger.info("(c) omega|ml by one2seven GmbH, Zurich, Switzerland, https://omegaml.io")
            logger.info("third party components (c) by their respective copyright holders")
            logger.info("see LICENSE and THIRDPARTY-LICENSES at https://github.com/omegaml/omegaml/")
            return
        self.parser.help()



from getpass import getuser
from unittest import TestCase

import pandas as pd
from pandas.util.testing import assert_frame_equal
from sqlalchemy.engine import Connection, create_engine, ResultProxy

from omegaml import Omega
from omegaml.backends.sqlalchemy import SQLAlchemyBackend
from omegaml.tests.util import OmegaTestMixin


class SQLAlchemyBackendTests(OmegaTestMixin, TestCase):
    def setUp(self):
        self.om = Omega()
        self.om.models.register_backend(SQLAlchemyBackend.KIND, SQLAlchemyBackend)
        self.clean()

    def test_put_connection(self):
        """
        store generic sqlalchemy connection
        """
        om = self.om
        cnx = 'sqlite:///test.db'
        om.datasets.put(cnx, 'testsqlite', kind=SQLAlchemyBackend.KIND)
        self.assertIn('testsqlite', om.datasets.list())
        meta = om.datasets.metadata('testsqlite')
        self.assertEqual(meta.kind, SQLAlchemyBackend.KIND)
        conn = om.datasets.get('testsqlite')
        self.assertIsInstance(conn, Connection)

    def test_put_connection_with_secrets(self):
        """
        store generic sqlalchemy connection
        """
        om = self.om
        cnx = 'sqlite:///{user}.db'
        om.datasets.put(cnx, 'testsqlite', kind=SQLAlchemyBackend.KIND)
        self.assertIn('testsqlite', om.datasets.list())
        meta = om.datasets.metadata('testsqlite')
        self.assertEqual(meta.kind, SQLAlchemyBackend.KIND)
        self.assertIn('secrets', meta.kind_meta)
        # no secrets
        with self.assertRaises(KeyError):
            om.datasets.get('testsqlite')
        # directly specified
        conn = om.datasets.get('testsqlite', secrets=dict(user='user'))
        # via vault
        om.datasets.put(dict(userid=getuser(), user='foobar'), '_omega/vault', append=False)
        conn = om.datasets.get('testsqlite')
        self.assertIsInstance(conn, Connection)

    def test_put_connection_with_sql(self):
        """
        store generic sqlalchemy connection with sql, same principle as a view
        """
        om = self.om
        cnx_str = 'sqlite:///test.db'
        engine = create_engine(cnx_str)
        cnx = engine.connect()
        df = pd.DataFrame({
            'x': range(10)
        })
        df.to_sql('foobar', cnx, if_exists='replace', index=False)
        om.datasets.put(cnx_str, 'foobar',
                        sql='select * from foobar',
                        kind=SQLAlchemyBackend.KIND)
        self.assertIn('foobar', om.datasets.list())
        meta = om.datasets.metadata('foobar')
        self.assertEqual(meta.kind, SQLAlchemyBackend.KIND)
        dfx = om.datasets.get('foobar')
        assert_frame_equal(dfx, df)

    def test_put_connection_with_sql_vars(self):
        """
        store generic sqlalchemy connection with sql containing variables
        """
        om = self.om
        cnx_str = 'sqlite:///test.db'
        engine = create_engine(cnx_str)
        cnx = engine.connect()
        df = pd.DataFrame({
            'x': range(10)
        })
        df.to_sql('foobar', cnx, if_exists='replace', index=False)
        om.datasets.put(cnx_str, 'foobar',
                        sql='select {cols} from foobar',
                        kind=SQLAlchemyBackend.KIND)
        self.assertIn('foobar', om.datasets.list())
        meta = om.datasets.metadata('foobar')
        self.assertEqual(meta.kind, SQLAlchemyBackend.KIND)
        with self.assertRaises(KeyError):
            # missing sqlvars
            dfx = om.datasets.get('foobar')
        dfx = om.datasets.get('foobar', sqlvars=dict(cols='x'))
        assert_frame_equal(dfx, df)

    def test_put_connection_with_sql_vars_lazy(self):
        """
        store generic sqlalchemy connection with sql containing variables
        """
        om = self.om
        cnx_str = 'sqlite:///test.db'
        engine = create_engine(cnx_str)
        cnx = engine.connect()
        df = pd.DataFrame({
            'x': range(10)
        })
        df.to_sql('foobar', cnx, if_exists='replace', index=False)
        om.datasets.put(cnx_str, 'foobar',
                        sql='select {cols} from foobar',
                        kind=SQLAlchemyBackend.KIND)
        self.assertIn('foobar', om.datasets.list())
        meta = om.datasets.metadata('foobar')
        self.assertEqual(meta.kind, SQLAlchemyBackend.KIND)
        with self.assertRaises(KeyError):
            # missing sqlvars
            dfx = om.datasets.get('foobar')
        dfx = om.datasets.getl('foobar', sqlvars=dict(cols='x'))
        self.assertIsInstance(dfx, ResultProxy)

    def test_put_copy_from_connection(self):
        """
        store generic sqlalchemy connection
        """
        om = self.om
        cnx_str = 'sqlite:///test.db'
        engine = create_engine(cnx_str)
        cnx = engine.connect()
        df = pd.DataFrame({
            'x': range(10)
        })
        df.to_sql('foobar', cnx, if_exists='replace', index=False)
        om.datasets.put(cnx_str, 'foobar_copy',
                        copy=True,
                        sql='select * from foobar',
                        kind=SQLAlchemyBackend.KIND)
        self.assertIn('foobar_copy', om.datasets.list())
        meta = om.datasets.metadata('foobar_copy')
        self.assertEqual(meta.kind, 'pandas.dfrows')
        dfx = om.datasets.get('foobar_copy')
        assert_frame_equal(dfx, df)

    def test_put_connection_with_sql_no_index(self):
        """
        store sql alchemy connection to specific query
        """
        # -- no index, will use default to_sql, which is 'index'
        df = pd.DataFrame({
            'x': range(10)
        })
        self._test_put_connection_with_sql_index(df, index_names=None)

    def test_put_connection_with_sql_default_index(self):
        """
        store store connection to specific query with default index name
        """
        # -- specify the default index explicitely
        df = pd.DataFrame({
            'x': range(10)
        })
        self._test_put_connection_with_sql_index(df, index_names=['index'])

    def test_put_connection_with_sql_custom_index(self):
        """
        store store connection to specific query with custom index name
        """
        # -- specify a custom index name
        df = pd.DataFrame({
            'x': range(10)
        })
        self._test_put_connection_with_sql_index(df, index_names=['idx'])

    def _test_put_connection_with_sql_index(self, df, index_names=None):
        om = self.om
        cnx_str = 'sqlite:///test.db'
        engine = create_engine(cnx_str)
        cnx = engine.connect()
        if index_names:
            df.index.names = index_names
        df.to_sql('foobar', cnx, if_exists='replace', index=index_names is not None)
        # store connection string
        sql = "select * from foobar"
        om.datasets.put(cnx_str, 'testsqlite', kind=SQLAlchemyBackend.KIND,
                        sql=sql, index_columns=index_names)
        self.assertIn('testsqlite', om.datasets.list())
        meta = om.datasets.metadata('testsqlite')
        self.assertEqual(meta.kind, SQLAlchemyBackend.KIND)
        # get back the connection only
        conn = om.datasets.get('testsqlite', raw=True)
        self.assertIsInstance(conn, Connection)
        # get back data using the stored sql
        dfx = om.datasets.get('testsqlite')
        assert_frame_equal(dfx, df)

    def test_put_data_via_connection(self):
        """
        store dataframe via connection
        """
        om = self.om
        cnx = 'sqlite:///test.db'
        om.datasets.put(cnx, 'testsqlite', table='foo',
                        kind=SQLAlchemyBackend.KIND)
        df = pd.DataFrame({
            'x': range(10)
        })
        df.index.names = ['index']
        # replace
        om.datasets.put(df, 'testsqlite', insert=True, append=False)
        meta = om.datasets.metadata('testsqlite')
        self.assertEqual(meta.kind, SQLAlchemyBackend.KIND)
        dfx = om.datasets.get('testsqlite', sql='select * from foo')
        assert_frame_equal(df, dfx)
        # append
        om.datasets.put(df, 'testsqlite', insert=True, append=False)
        om.datasets.put(df, 'testsqlite', insert=True)
        dfx = om.datasets.get('testsqlite', sql='select * from foo')
        df_expected = df.append(df)
        assert_frame_equal(dfx, df_expected)

    def test_put_raw_data_via_connection(self):
        """
        store raw dict via connection
        """
        om = self.om
        # store connection
        cnx = 'sqlite:///test.db'
        om.datasets.put(cnx, 'testsqlite', table='foo',
                        kind=SQLAlchemyBackend.KIND)
        df = pd.DataFrame({
            'x': range(10)
        })
        df.index.names = ['index']
        raw = df.to_dict('dict')
        # replace
        om.datasets.put(raw, 'testsqlite', insert=True, append=False)
        meta = om.datasets.metadata('testsqlite')
        self.assertEqual(meta.kind, SQLAlchemyBackend.KIND)
        dfx = om.datasets.get('testsqlite', sql='select * from foo')
        assert_frame_equal(df, dfx)
        # append
        om.datasets.put(raw, 'testsqlite', insert=True, append=False)
        om.datasets.put(raw, 'testsqlite', insert=True)
        dfx = om.datasets.get('testsqlite', sql='select * from foo')
        df_expected = df.append(df)
        assert_frame_equal(dfx, df_expected)



def protected(kw):
    return '__' + kw


def get_omega(args):
    from omegaml import setup, _base_config
    from omegaml.client.cloud import setup_from_config
    config_file = args.get('--config')
    local_runtime = args.get('--local-runtime')
    if local_runtime:
        _base_config.OMEGA_LOCAL_RUNTIME = True
    bucket = args.get('--bucket')
    if config_file:
        om = setup_from_config(config_file)
    else:
        om = setup()
    return om[bucket]



from __future__ import absolute_import

import datetime
from mongoengine.base.fields import ObjectIdField
from mongoengine.document import Document
from mongoengine.fields import (
    StringField, FileField, DictField, DateTimeField
)

from omegaml.util import settings


# default kinds of objects
class MDREGISTRY:
    PANDAS_DFROWS = 'pandas.dfrows'  # dataframe
    PANDAS_SEROWS = 'pandas.serows'  # series
    PANDAS_HDF = 'pandas.hdf'
    PYTHON_DATA = 'python.data'
    PANDAS_DFGROUP = 'pandas.dfgroup'
    SKLEARN_JOBLIB = 'sklearn.joblib'
    OMEGAML_JOBS = 'script.ipynb'
    SPARK_MLLIB = 'spark.mllib'
    OMEGAML_RUNNING_JOBS = 'job.run'

    #: the list of accepted data types. extend using OmegaStore.register_backend
    KINDS = [
        PANDAS_DFROWS, PANDAS_SEROWS, PANDAS_HDF, PYTHON_DATA, SKLEARN_JOBLIB,
        PANDAS_DFGROUP, OMEGAML_JOBS, OMEGAML_RUNNING_JOBS, SPARK_MLLIB
    ]


class Metadata:
    """
    Metadata stores information about objects in OmegaStore
    """

    # NOTE THIS IS ONLY HERE FOR DOCUMENTATION PURPOSE.
    #
    # If you use this class to save a document, it will raise a NameError
    #
    # The actual Metadata class is created in make_Metadata() below.
    # Rationale: If we let mongoengine create Metadata here the class
    # is bound to a specific MongoClient instance. Using make_Metadata
    # binds the class to the specific instance that exists at the time
    # of creation. Open to better ways.

    # fields
    #: this is the name of the data
    name = StringField(unique_with=['bucket', 'prefix'])
    #: bucket
    bucket = StringField()
    #: prefix
    prefix = StringField()
    #: kind of data
    kind = StringField(choices=MDREGISTRY.KINDS)
    #: for PANDAS_HDF and SKLEARN_JOBLIB this is the gridfile
    gridfile = FileField(
        db_alias='omega',
        collection_name=settings().OMEGA_MONGO_COLLECTION)
    #: for PANDAS_DFROWS this is the collection
    collection = StringField()
    #: for PYTHON_DATA this is the actual document
    objid = ObjectIdField()
    #: omegaml technical attributes, e.g. column indicies
    kind_meta = DictField()
    #: customer-defined other meta attributes
    attributes = DictField()
    #: s3file attributes
    s3file = DictField()
    #: location URI
    uri = StringField()
    #: created datetime
    created = DateTimeField(default=datetime.datetime.now)
    #: created datetime
    modified = DateTimeField(default=datetime.datetime.now)


def make_Metadata():
    # this is to create context specific Metadata class that takes the
    # database from the given alias at the time of use
    from omegaml.documents import Metadata as Metadata_base
    class Metadata(Metadata_base, Document):
        # the actual db is defined at runtime
        meta = {
            'db_alias': 'omega',
            'strict': False,
            'indexes': [
                # unique entry
                {
                    'fields': ['bucket', 'prefix', 'name'],
                },
                'created',  # most recent is last, i.e. [-1]
            ]
        }

        def __new__(cls, *args, **kwargs):
            # undo the Metadata.__new__ protection
            return super(Metadata, cls).__real_new__(cls)

        def __eq__(self, other):
            return self.objid == other.objid

        def __unicode__(self):
            fields = ('name', 'bucket', 'prefix', 'created', 'kind')
            kwargs = ('%s=%s' % (k, getattr(self, k))
                      for k in self._fields.keys() if k in fields)
            return u"Metadata(%s)" % ','.join(kwargs)

        def save(self, *args, **kwargs):
            assert self.name is not None, "a dataset name is needed before saving"
            self.modified = datetime.datetime.now()
            return super(Metadata_base, self).save(*args, **kwargs)

    return Metadata


def make_QueryCache():
    class QueryCache(Document):
        collection = StringField()
        key = StringField()
        value = DictField()
        meta = {
            'db_alias': 'omega',
            'indexes': [
                'key',
            ]
        }

    return QueryCache


def raise_on_use(exc):
    def inner(*args, **kwargs):
        raise exc

    return inner


Metadata.__real_new__ = Metadata.__new__
Metadata.__new__ = raise_on_use(NameError("You must use make_Metadata()() to instantiate a working object"))



from __future__ import absolute_import

import logging

from celery.exceptions import NotRegistered

logger = logging.getLogger(__file__)


class OmegaScriptProxy(object):
    """
    proxy to a remote job in a celery worker

    Usage:

        .. code::

            om = Omega()
            # result is AsyncResult, use .get() to return it's result
            result = om.runtimes.script('foojob').run()
            result.get()
    """

    def __init__(self, scriptname, runtime=None):
        self.scriptname = scriptname
        self.runtime = runtime

    def run(self, **kwargs):
        """
        run the script

        Runs the script and returns its result as

        {"runtimes": 7.4e-05,
          "started": "2018-04-07T17:57:52.451012",
          "kwargs": {},
          "result": {},
          "script": "helloworld"
        }

        :return: the result
        """
        script_run = self.runtime.task('omegaml.backends.package.tasks.run_omega_script')
        return script_run.delay(self.scriptname, **kwargs)



from omegaml.client.docoptparser import CommandBase
from omegaml.client.util import get_omega


class RuntimeCommandBase(CommandBase):
    """
    Usage:
      om runtime model <name> <model-action> [<X>] [<Y>] [--result=<output-name>] [--param=<kw=value>]... [options]
      om runtime script <name> [<script-action>] [<kw=value>...] [--async] [options]
      om runtime job <name> [<job-action>] [<args...>] [--async] [options]
      om runtime result <taskid> [options]
      om runtime ping [options]
      om runtime log [-f]

    Options:
      --async  don't wait for results, will print taskid
      -f       tail log

    Description:
      <model-action> can be any valid model action like fit, predict, score,
      transform, decision_function etc.

      <script-action> defaults to run
      <job-action> defaults to run

      Examples:
        om runtime model <name> fit <X> <Y>
        om runtime model <name> predict <X>
    """
    command = 'runtime'

    def ping(self):
        om = get_omega(self.args)
        self.logger.info(om.runtime.ping())

    def _ensure_valid_XY(self, value):
        if value is not None:
            if value.isnumeric():
                return [eval(value)]
            if value[0] == '[' and value[-1] == ']':
                return eval(value)
        return value

    def model(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        action = self.args.get('<model-action>')
        is_async = self.args.get('--async')
        kwargs_lst = self.args.get('--param')
        output = self.args.get('--result')
        X = self._ensure_valid_XY(self.args.get('<X>'))
        Y = self._ensure_valid_XY(self.args.get('<Y>'))
        # parse the list of kw=value values
        # e.g. key1=val1 key2=val2 => kwargs_lst = ['key1=val1', 'key2=val2']
        #   => kw_dct = { 'key1': eval('val1'), 'key2': eval('val2') }
        kv_dct = {}
        for kv in kwargs_lst:
            k, v = kv.split('=', 1)
            kv_dct[k] = eval(v)
        kwargs = {}
        if action in ('predict', 'predict_proba',
                      'decision_function', 'transform', 'score'):
            # actions that take rName, but no Y
            kwargs['rName'] = output
        else:
            # actions that take Y, but no rName
            kwargs['Yname'] = Y
        if action == 'gridsearch':
            kwargs['parameters'] = kv_dct
        rt_model = om.runtime.model(name)
        meth = getattr(rt_model, action, None)
        if meth is not None:
            result = meth(X, **kwargs)
            if not is_async:
                self.logger.info(result.get())
            else:
                self.logger.info(result)
            return
        raise ValueError('{action} is not applicable to {name}'.format(**locals()))

    def script(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        is_async = self.args.get('--async')
        kwargs = self.parse_kwargs('<kw=value>')
        result = om.runtime.script(name).run(**kwargs)
        if not is_async:
            self.logger.info(result.get())
        else:
            self.logger.info(result.task_id)

    def job(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        is_async = self.args.get('--async')
        result = om.runtime.job(name).run()
        if not is_async:
            self.logger.info(result.get())
        else:
            self.logger.info(result.task_id)

    def result(self):
        from celery.result import AsyncResult

        om = get_omega(self.args)
        task_id = self.args.get('<taskid>')
        result = AsyncResult(task_id, app=om.runtime.celeryapp).get()
        self.logger.info(result)

    def log(self):
        import pandas as pd
        tail  = self.args.get('-f')
        om = get_omega(self.args)
        if not tail:
            df = om.logger.dataset.get()
            with pd.option_context('display.max_rows', None,
                                   'display.max_columns', None,
                                   'display.max_colwidth', -1):
                print(df[['text']])
        else:
            om.logger.dataset.tail()



from unittest.case import TestCase

import omegaml as om
from omegaml.backends.basemodel import BaseModelBackend
from omegaml.documents import MDREGISTRY
from omegaml.backends.basedata import BaseDataBackend


class CustomBackendTests(TestCase):

    """
    Test custom backends with new kinds

    Note we are not implemented actually working backends. What we're
    interested here is that the backend API and Metadata storage work.
    """
    def tearDown(self):
        # remove custom backend from implementation not to disturb other tests
        try:
            del om.defaults.OMEGA_STORE_BACKENDS['custom.foo']
            del om.defaults.OMEGA_STORE_BACKENDS['custom.bar']
            MDREGISTRY.KINDS.remove('custom.foo')
            MDREGISTRY.KINDS.remove('custom.bar')
        except:
            pass

    def test_custom_model_backend(self):
        """
        test custom model type 
        """
        om.models.register_backend('custom.foo', CustomModelBackend)
        foo = dict(foo='bar')
        meta = om.models.put(foo, 'footest')
        self.assertIsInstance(meta, om.models._Metadata)
        self.assertEqual(meta.kind, 'custom.foo')
        meta_stored = om.models.metadata('footest')
        self.assertIn('footest', om.models.list())
        self.assertEqual(meta, meta_stored)
        with self.assertRaises(NotImplementedError):
            om.models.get('footest')

    def test_custom_dataset_backend(self):
        """
        test custom dataset type 
        """
        om.datasets.register_backend('custom.bar', CustomDataBackend)
        foo = dict(bar='foo')
        meta = om.datasets.put(foo, 'bartest')
        self.assertIsInstance(meta, om.datasets._Metadata)
        self.assertEqual(meta.kind, 'custom.bar')
        meta_stored = om.datasets.metadata('bartest')
        self.assertIn('bartest', om.datasets.list())
        self.assertEqual(meta, meta_stored)
        with self.assertRaises(NotImplementedError):
            om.datasets.get('bartest')


class CustomModelBackend(BaseModelBackend):

    """
    Minimalist model backend
    """

    @classmethod
    def supports(self, obj, name, **kwargs):
        return isinstance(obj, dict) and 'foo' in obj

    def put_model(self, obj, name, attributes=None):
        kind = 'custom.foo'
        return self.model_store.make_metadata(name, kind, bucket=None,
                                              prefix=None,
                                              attributes=attributes).save()


class CustomDataBackend(BaseDataBackend):

    """
    Minimalist dataset backend
    """

    @classmethod
    def supports(self, obj, name, **kwargs):
        return isinstance(obj, dict) and 'bar' in obj

    def put(self, obj, name, attributes=None):
        kind = 'custom.bar'
        return self.model_store.make_metadata(name, kind, bucket=None,
                                              prefix=None,
                                              attributes=attributes).save()



import sys

import yaml

from omegaml.client.auth import OmegaRestApiAuth, OmegaRuntimeAuthentication


def get_user_config_from_api(api_auth, api_url=None, requested_userid=None, view=False):
    # safe way to talk to either the remote API or the in-process test server
    from omegaml import settings
    defaults = settings()
    api_url = api_url or defaults.OMEGA_RESTAPI_URL
    api_url += '/api/v1/config/'
    api_url = api_url.replace('//api', '/api')
    query = []
    if requested_userid:
        query.append('user={}'.format(requested_userid))
    if view:
        query.append('view={}'.format(int(view)))
    api_url += '?' + '&'.join(query)
    # -- setup appropriate client API
    if api_url.startswith('http'):
        import requests
        server = requests
        server_kwargs = dict(auth=api_auth)
        deserialize = lambda resp: resp.json()
    elif any('test' in v for v in sys.argv):
        # test support
        import json
        from tastypie.test import TestApiClient
        server = TestApiClient()
        server_kwargs = dict(authentication=api_auth.get_credentials())
        deserialize = lambda resp: json.loads(resp.content.decode('utf-8'))
    else:
        raise ValueError('invalid api_url {}'.format(api_url))
    # -- actual logic to get configs
    fail_msg = ("Not authenticated using userid {api_auth.username}"
                " apikey {api_auth.apikey}, error was {resp.status_code} "
                "using {api_url}\n{resp.content}")
    resp = server.get(api_url, **server_kwargs)
    assert resp.status_code == 200, fail_msg.format(**locals())
    configs = deserialize(resp)
    return configs


# FIXME enable cache by arguments (mnemonic) @cached(seconds=3600)
def get_omega_from_apikey(userid, apikey, api_url=None, requested_userid=None,
                          qualifier=None, view=False):
    """
    setup an Omega instance from userid and apikey

    :param userid: the userid
    :param apikey: the apikey
    :param api_url: the api URL
    :param requested_userid: the userid to request config for. in this case userid
      and apikey must for a staff user for the request to succeed
    :param qualifier: the database qualifier requested. defaults to 'default'
    :returns: the Omega instance configured for the given user
    """
    from omegaml import Omega
    from omegaml import settings, _base_config
    from omegaml.util import DefaultsContext

    defaults = DefaultsContext(settings())
    qualifier = qualifier or 'default'
    api_url = api_url or defaults.OMEGA_RESTAPI_URL
    if api_url.startswith('http') or any('test' in v for v in sys.argv):
        api_auth = OmegaRestApiAuth(userid, apikey)
        configs = get_user_config_from_api(api_auth, api_url=api_url,
                                           requested_userid=requested_userid,
                                           view=view)
        configs = configs['objects'][0]['data']
    elif api_url == 'local':
        configs = {k: getattr(defaults, k) for k in dir(defaults) if k.startswith('OMEGA')}
    else:
        raise ValueError('invalid api_url {}'.format(api_url))
    if qualifier == 'default':
        config = configs.get(qualifier, configs)
    else:
        config = configs[qualifier]
    _base_config.update_from_dict(config, attrs=defaults)
    auth = OmegaRuntimeAuthentication(userid, apikey, qualifier)
    om = Omega(defaults=defaults, auth=auth)
    # update config to reflect request
    om.defaults.OMEGA_RESTAPI_URL = api_url
    om.defaults.OMEGA_USERID = userid
    om.defaults.OMEGA_APIKEY = apikey
    return om


def get_omega_from_config(configfile, qualifier=None):
    from omegaml import Omega
    from omegaml import settings, _base_config
    defaults = settings()
    with open(configfile, 'r') as fconfig:
        configs = yaml.safe_load(fconfig)
    qualifier = qualifier or 'default'
    if qualifier == 'default':
        config = configs.get(qualifier, configs)
    else:
        config = configs[qualifier]
    _base_config.update_from_dict(config, attrs=defaults)
    settings(reload=True)
    om = Omega(defaults=defaults)
    return om


def save_userconfig_from_apikey(configfile, userid, apikey, api_url=None, requested_userid=None,
                                view=False):
    from omegaml import settings
    defaults = settings()
    api_url = api_url or defaults.OMEGA_RESTAPI_URL
    with open(configfile, 'w') as fconfig:
        auth = OmegaRestApiAuth(userid, apikey)
        configs = get_user_config_from_api(auth,
                                           api_url=api_url,
                                           requested_userid=requested_userid,
                                           view=view)
        config = configs['objects'][0]['data']
        config['OMEGA_RESTAPI_URL'] = api_url
        yaml.safe_dump(config, fconfig, default_flow_style=False)
        print("Config is in {configfile}".format(**locals()))



import tempfile

from omegaml.backends.tensorflow import TensorflowSavedModelBackend


class TensorflowKerasSavedModelBackend(TensorflowSavedModelBackend):
    KIND = 'tfkeras.savedmodel'

    @classmethod
    def supports(self, obj, name, **kwargs):
        import tensorflow as tf
        tfSequential = tf.keras.models.Sequential
        tfModel = tf.keras.models.Model
        return isinstance(obj, (tfSequential, tfModel)) and kwargs.get('as_savedmodel')

    def _make_savedmodel(self, obj, serving_input_receiver_fn=None, strip_default_attrs=None):
        # https://www.tensorflow.org/api_docs/python/tf/keras/experimental/export_saved_model
        import tensorflow as tf
        export_dir = tempfile.mkdtemp()
        tf.contrib.saved_model.save_keras_model(obj, export_dir,
                                                input_signature=[tf.TensorSpec(shape=obj.input.shape,
                                                                               dtype=obj.input.dtype,
                                                                               name='input')])
        return export_dir

    def fit(self, modelname, Xname, Yname=None, pure_python=True, tpu_specs=None, **kwargs):
        raise ValueError('cannot fit a saved model')



from unittest import TestCase

import os

import nbformat

from omegaml import Omega
from omegaml.tests.cli.scenarios import CliTestScenarios
from omegaml.tests.util import OmegaTestMixin
from omegaml.util import temp_filename


class CliScriptsTest(CliTestScenarios, OmegaTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.om = Omega()
        self.clean()

    def test_cli_scripts_list_put(self):
        self.cli(f'scripts list')
        self.assertLogContains('info', '[]')
        pkgpath = self.get_package_path()
        self.cli(f'scripts put {pkgpath} helloworld', new_start=True)
        self.assertLogContains('info', 'Metadata')
        self.assertLogContains('info', 'name=helloworld')
        self.cli(f'scripts list')
        self.assertLogSize('info', 1)
        self.assertLogContains('info', 'helloworld')

    def test_cli_scripts_drop(self):
        pkgpath = self.get_package_path()
        self.cli(f'scripts put {pkgpath} helloworld', new_start=True)
        self.assertIn('helloworld', self.om.scripts.list())
        self.cli(f'scripts delete helloworld', new_start=True)
        self.assertNotIn('helloworld', self.om.scripts.list())

